#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""The Cache subclass module."""
from __future__ import annotations

import contextlib
import fnmatch
import sys
from typing import TYPE_CHECKING, Generator

import apt_pkg
from apt.cache import Cache as _Cache
from apt.package import Package

from nala import _, color, color_version
from nala.constants import ERROR_PREFIX, NOTICE_PREFIX, WARNING_PREFIX
from nala.options import arguments
from nala.rich import Columns, from_ansi
from nala.utils import dprint, eprint, term

if TYPE_CHECKING:
	from nala.debfile import NalaDebPackage
	from nala.dpkg import InstallProgress, UpdateProgress


PACKAGES_CAN_BE_UPGRADED = "\n" + _("The following {total} packages can be upgraded:")
NOT_CANDIDATE = color("[") + color(_("Not candidate version"), "YELLOW") + color("]")


class Cache(_Cache):
	"""Subclass of apt.cache to add features."""

	def commit_pkgs(
		self,
		install_progress: InstallProgress,
		update_progress: UpdateProgress,
		local_debs: list[NalaDebPackage] | None = None,
	) -> bool:
		"""Apply the marked changes to the cache."""
		if local_debs:
			res = install_archives(
				[pkg.filename for pkg in local_debs if pkg.filename], install_progress
			)
			install_progress.finish_update()
			return res == 0
		with apt_pkg.SystemLock():
			apt = apt_pkg.PackageManager(self._depcache)
			fetcher = apt_pkg.Acquire(update_progress)
			with self._archive_lock:
				while True:
					self._fetch_archives(fetcher, apt)
					res = install_archives(apt, install_progress)
					if res == apt.RESULT_COMPLETED:
						break
					if res == apt.RESULT_FAILED:
						raise SystemError("installArchives() failed")
					if res != apt.RESULT_INCOMPLETE:
						raise SystemError(
							f"internal-error: unknown result code from InstallArchives: {res}"
						)
					# reload the fetcher for media swapping
					fetcher.shutdown()
		return res == apt.RESULT_COMPLETED

	def is_secret_virtual(self, pkg_name: str) -> bool:
		"""Return True if the package is secret virtual.

		Secret virtual packages provide nothing, and have no versions.

		cache.is_virtual_package() returns True only if the virtual package
		has something that it can provide.
		"""
		try:
			pkg = self._cache[pkg_name]
			return not pkg.has_provides and not pkg.has_versions
		except KeyError:
			return False

	def is_any_virtual(self, pkgname: str) -> bool:
		"""Return whether the package is a virtual package.

		This is used if we only care if it's virtual, but not what type.
		"""
		try:
			pkg = self._cache[pkgname]
		except KeyError:
			return False
		return not pkg.has_versions

	def glob_filter(self, pkg_names: list[str], show: bool = False) -> list[str]:
		"""Filter provided packages and glob *.

		Returns a new list of packages matching the glob.

		If there is nothing to glob it returns the original list.
		"""
		if "*" not in f"{pkg_names}":
			return pkg_names

		new_packages: list[str] = []
		glob_failed = False
		for pkg_name in pkg_names:
			if "*" in pkg_name:
				dprint(f"Globbing: {pkg_name}")
				glob = fnmatch.filter(self.get_pkg_names(show), pkg_name)
				if not glob:
					glob_failed = True
					eprint(
						_(
							"{error} unable to find any packages by globbing {pkg}"
						).format(error=ERROR_PREFIX, pkg=color(pkg_name, "YELLOW"))
					)
					continue
				new_packages.extend(glob)
			else:
				new_packages.append(pkg_name)

		if glob_failed:
			sys.exit(1)
		new_packages.sort()
		dprint(f"List after globbing: {new_packages}")
		return new_packages

	def get_pkg_names(self, show: bool = False) -> Generator[str, None, None]:
		"""Generate all real packages, or packages that can provide something."""
		for pkg in self._cache.packages:  # pylint: disable=not-an-iterable
			pretty_name = pkg.get_fullname(pretty=True)
			if not show and self.is_virtual_package(pretty_name):
				provides = self.get_providing_packages(pretty_name)
				if not provides or len(provides) > 1:
					continue
			# For some reason a virtual package $kernel exists and can't be accessed.
			if pretty_name.startswith("$"):
				continue
			if pkg.has_versions or pkg.has_provides:
				yield pretty_name

	def virtual_filter(self, pkg_names: list[str], remove: bool = False) -> list[str]:
		"""Filter package to check if they're virtual."""
		new_names = set()
		for pkg_name in pkg_names:
			if pkg_name in self:
				new_names.add(pkg_name)
				continue
			if (vpkg := self.check_virtual(pkg_name, remove)) and isinstance(
				vpkg, Package
			):
				new_names.add(vpkg.name)
				continue
			new_names.add(pkg_name)
		dprint(f"Virtual Filter: {new_names}")
		return sorted(new_names)

	def what_replaces(self, pkg_name: str) -> Generator[str, None, None]:
		"""Generate packages that replace the given name."""
		for pkg in self._cache.packages:  # pylint: disable=not-an-iterable
			if cand := self._depcache.get_candidate_ver(pkg):
				with contextlib.suppress(KeyError):
					replaces = cand.depends_list["Replaces"]
					target = replaces[0][0].target_pkg
					if pkg_name == target.name:
						yield pkg.get_fullname(pretty=True)

	def check_virtual(self, pkg_name: str, remove: bool = False) -> Package | bool:
		"""Check if the package is virtual."""
		if not self.is_virtual_package(pkg_name):
			return False
		if len(provides := self.get_providing_packages(pkg_name)) == 1:
			print_selecting_pkg(provides[0].name, pkg_name)
			return self[provides[0]]
		if remove:
			eprint(
				_("{warn} Virtual Packages like {package} can't be removed.").format(
					warn=WARNING_PREFIX, package=color(pkg_name, "YELLOW")
				)
			)
			return True
		if not provides:
			if provides := self.get_providing_packages(pkg_name, candidate_only=False):
				print_virtual_pkg(pkg_name, provides, not_candidate=True)
				return True
		print_virtual_pkg(pkg_name, provides)
		return True

	def purge_removed(self) -> None:
		"""Make sure everything marked as removed is getting purged."""
		if not arguments.is_purge():
			return
		for pkg in self:
			if pkg.marked_delete:
				pkg.mark_delete(purge=True)

	def protect_upgrade_pkgs(self, exclude: list[str] | None) -> set[Package]:
		"""Mark excluded packages as protected."""
		protected: set[Package] = set()
		if not exclude:
			return protected
		resolver = apt_pkg.ProblemResolver(self._depcache)
		for pkg_name in self.glob_filter(exclude):
			if pkg_name in self:
				pkg = self[pkg_name]
				if pkg.is_upgradable:
					print(
						_("Protecting {package} from upgrade").format(
							package=color(pkg_name, "GREEN")
						)
					)
					resolver.protect(self._cache[pkg_name])
					protected.add(pkg)
				elif pkg.is_auto_removable:
					print(
						_("Protecting {package} from auto-removal").format(
							package=color(pkg_name, "GREEN")
						)
					)
					resolver.protect(self._cache[pkg_name])
					protected.add(pkg)
		return protected

	def upgradable_pkgs(self) -> Generator[Package, None, None]:
		"""Generate upgradable packages."""
		return (pkg for pkg in self if pkg.is_upgradable)

	def print_upgradable(self) -> None:
		"""Print packages that are upgradable."""
		if arguments.config.get_bool("update_show_packages"):
			if upgradable := [
				# format will look like "python3-pip (22.1.1+dfsg-1) -> (22.2+dfsg-1)"
				from_ansi(
					f"{color(pkg.name, 'GREEN')} "
					f"{color_version(pkg.installed.version)} -> {color_version(pkg.candidate.version)}"
				)
				for pkg in self.upgradable_pkgs()
				if pkg.installed and pkg.candidate
			]:
				print(PACKAGES_CAN_BE_UPGRADED.format(total=color(len(upgradable))))
				term.console.print(Columns(upgradable, padding=(0, 2), equal=True))
				return

		elif total_pkgs := len(tuple(self.upgradable_pkgs())):
			print(
				_(
					"{total} packages can be upgraded. Run '{command}' to see them."
				).format(
					total=color(total_pkgs, "YELLOW"),
					command=color("nala list --upgradable", "GREEN"),
				)
			)
			return

		print(color(_("All packages are up to date.")))


def install_archives(
	apt: apt_pkg.PackageManager | list[str], install_progress: InstallProgress
) -> int:
	"""Install the archives."""
	install_progress.start_update()

	if did_unlock := apt_pkg.pkgsystem_is_locked():
		apt_pkg.pkgsystem_unlock_inner()

	try:
		res = install_progress.run_install(apt)
	finally:
		if did_unlock:
			apt_pkg.pkgsystem_lock_inner()

	install_progress.finish_update()
	return res


def print_virtual_pkg(
	pkg_name: str, provides: list[Package], not_candidate: bool = False
) -> None:
	"""Print the virtual package string."""
	print(
		_("{package} is a virtual package provided by:").format(
			package=color(pkg_name, "GREEN")
		)
	)
	print(
		"".join(
			[
				f"\n  {color(pkg.name, 'GREEN')} {color_version(pkg.candidate.version)} "
				f"{NOT_CANDIDATE if not_candidate else ''}"
				for pkg in provides
				if pkg.candidate
			]
		).strip("\n")
	)
	print(_("You should select just one."))


def print_selecting_pkg(provider: str, pkg_name: str) -> None:
	"""Print that we are selecting a different package."""
	print(
		_(
			"{notice} Selecting {provider}\n  Instead of virtual package {package}"
		).format(
			notice=NOTICE_PREFIX,
			provider=color(provider, "GREEN"),
			package=color(pkg_name, "GREEN"),
		),
		end="\n\n",
	)



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
# nala is based upon apt-metalink https://github.com/tatsuhiro-t/apt-metalink
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Main module for Nala which facilitates apt."""
from __future__ import annotations

import fnmatch
import re
import sys
from subprocess import CalledProcessError, run
from typing import Generator, List, Optional, Pattern, cast

import apt_pkg
import typer
from apt import Package, Version

from nala import _, color
from nala.cache import Cache
from nala.constants import (
	ARCHIVE_DIR,
	CAT,
	DPKG_STATE,
	ERROR_PREFIX,
	LISTS_DIR,
	LISTS_PARTIAL_DIR,
	NALA_SOURCES,
	PARTIAL_DIR,
	PKGCACHE,
	SRCPKGCACHE,
)
from nala.error import BrokenError, pkg_error
from nala.history import get_history, get_list
from nala.install import (
	auto_remover,
	check_broken,
	check_state,
	fix_excluded,
	get_changes,
	install_local,
	package_manager,
	setup_cache,
	split_local,
)
from nala.options import (
	ALL_ARCHES,
	ALL_VERSIONS,
	ASSUME_YES,
	AUTO_REMOVE,
	COLOR,
	CONFIG,
	DEBUG,
	DEFAULT_RELEASE,
	DOWNLOAD_ONLY,
	FETCH,
	FIX_BROKEN,
	FULL,
	FULL_UPGRADE,
	INSTALLED,
	LISTS,
	MAN_HELP,
	NALA_INSTALLED,
	NAMES,
	OPTION,
	PURGE,
	RAW_DPKG,
	RECOMMENDS,
	REMOVE_ESSENTIAL,
	SIMPLE,
	SUGGESTS,
	UPDATE,
	UPGRADABLE,
	UPGRADEABLE,
	VERBOSE,
	VIRTUAL,
	arguments,
	nala,
)
from nala.rich import ELLIPSIS
from nala.search import iter_search, search_name, skip_pkg
from nala.show import additional_notice, pkg_not_found, show_main
from nala.utils import (
	NalaPackage,
	PackageHandler,
	ask,
	command_help,
	compile_regex,
	dedupe_list,
	eprint,
	iter_remove,
	sudo_check,
	vprint,
)

# pylint: disable=unused-argument,too-many-arguments, too-many-locals

nala_pkgs = PackageHandler()


def _fix_broken(nested_cache: Cache | None = None) -> None:
	"""Attempt to fix broken packages, if any."""
	cache = nested_cache or setup_cache()
	print("Fixing Broken Packages...")
	cache.fix_broken()

	if nested_cache:
		print(color(_("There are broken packages that need to be fixed!"), "YELLOW"))
		print(
			_("You can use {switch} if you'd like to try without fixing them.").format(
				switch=color("--no-fix-broken", "YELLOW")
			)
		)
	else:
		check_state(cache, nala_pkgs)
	get_changes(cache, nala_pkgs, "fix-broken")


def _remove(pkg_names: list[str]) -> None:
	sudo_check()

	cache = setup_cache()
	check_state(cache, nala_pkgs)

	pkg_names = cache.glob_filter(dedupe_list(pkg_names))
	pkg_names = cache.virtual_filter(pkg_names, remove=True)
	broken, not_found, ver_failed = check_broken(
		pkg_names,
		cache,
		remove=True,
	)

	for pkg_name in not_found[:]:
		if cache.is_any_virtual(pkg_name):
			not_found.remove(pkg_name)
			pkg_names.remove(pkg_name)

	if not_found or ver_failed:
		pkg_error(not_found, cache)

	nala_pkgs.user_explicit = [cache[pkg_name] for pkg_name in pkg_names]
	if not package_manager(pkg_names, cache, remove=True):
		BrokenError(cache).broken_remove(broken)

	auto_remover(cache, nala_pkgs)
	get_changes(cache, nala_pkgs, "remove")


def _install(pkg_names: list[str] | None, ctx: typer.Context) -> None:
	sudo_check(pkg_names)
	if not pkg_names:
		if arguments.fix_broken:
			_fix_broken()
			return
		ctx.fail(_("{error} Missing packages to install").format(error=ERROR_PREFIX))

	pkg_names = dedupe_list(pkg_names)  # type: ignore[arg-type]
	check_state(cache := setup_cache(), nala_pkgs)
	not_exist = split_local(pkg_names, cache, nala_pkgs.local_debs)
	install_local(nala_pkgs, cache)

	pkg_names = cache.glob_filter(pkg_names)
	pkg_names = cache.virtual_filter(pkg_names)
	broken, not_found, ver_failed = check_broken(pkg_names, cache)
	not_found.extend(not_exist)

	if not_found or ver_failed:
		pkg_error(not_found, cache)

	nala_pkgs.user_explicit = [cache[pkg_name] for pkg_name in pkg_names]
	if not package_manager(pkg_names, cache):
		if not (error := BrokenError(cache, broken)).broken_install():
			error.unmarked_error(nala_pkgs.user_explicit)

	auto_remover(cache, nala_pkgs)
	get_changes(cache, nala_pkgs, "install")


def remove_completion(ctx: typer.Context) -> Generator[str, None, None]:
	"""Complete remove command arguments."""
	if not DPKG_STATE.exists():
		return
	regex = r"ok installed|half-installed|unpacked|half-configured"
	if "purge" in ctx.command_path:
		regex += r"|config-files"

	status = re.compile(regex)
	for package in DPKG_STATE.read_text(encoding="utf-8").split("\n\n"):
		pkg_name = pkg_status = None
		for line in package.splitlines():
			if len(fields := line.split(": ")) == 1:
				continue
			if "Package" in fields:
				pkg_name = fields[1]
			if "Status" in fields and status.findall(fields[1]):
				pkg_status = fields[1]
		if pkg_name and pkg_status:
			yield pkg_name


def package_completion(cur: str) -> Generator[str, None, None]:
	"""Complete install command arguments."""
	try:
		yield from run(
			["apt-cache", "--no-generate", "pkgnames", cur]
			if PKGCACHE.exists()
			else ["apt-cache", "pkgnames", cur],
			capture_output=True,
			check=True,
			text=True,
		).stdout.split()
	except CalledProcessError as error:
		sys.exit(f"\n{error.stderr}")


@nala.command("update", help=_("Update package list."))
def _update(
	debug: bool = DEBUG,
	raw_dpkg: bool = RAW_DPKG,
	dpkg_option: List[str] = OPTION,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Update package list."""
	sudo_check()
	arguments.update = True
	setup_cache().print_upgradable()


@nala.command("dist-upgrade", hidden=True)
@nala.command(
	"full-upgrade",
	help=_("Upgrade the system by removing/installing/upgrading packages."),
)
@nala.command(help=_("Upgrade the system by upgrading packages only."))
def upgrade(
	exclude: Optional[List[str]] = typer.Option(
		None,
		metavar="PKG",
		help=_("Specify packages to exclude during upgrade. Accepts glob*"),
	),
	purge: bool = PURGE,
	debug: bool = DEBUG,
	raw_dpkg: bool = RAW_DPKG,
	download_only: bool = DOWNLOAD_ONLY,
	remove_essential: bool = REMOVE_ESSENTIAL,
	full_upgrade: bool = FULL_UPGRADE,
	update: bool = UPDATE,
	auto_remove: bool = AUTO_REMOVE,
	install_recommends: bool = RECOMMENDS,
	install_suggests: bool = SUGGESTS,
	fix_broken: bool = FIX_BROKEN,
	assume_yes: bool = ASSUME_YES,
	dpkg_option: List[str] = OPTION,
	simple: bool = SIMPLE,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Update package list and upgrade the system."""
	sudo_check()

	# If alias is full-upgrade set --full to True
	if arguments.command in {"full-upgrade", "dist-upgrade"}:
		arguments.full_upgrade = True

	def _upgrade(
		exclude: list[str] | None = None,
		nested_cache: Cache | None = None,
	) -> None:
		"""Upgrade pkg[s]."""
		cache = nested_cache or setup_cache()
		check_state(cache, nala_pkgs)

		is_upgrade = tuple(cache.upgradable_pkgs())
		protected = cache.protect_upgrade_pkgs(exclude)
		try:
			cache.upgrade(dist_upgrade=arguments.full_upgrade)
		except apt_pkg.Error:
			if exclude:
				exclude = fix_excluded(protected, is_upgrade)
				if ask(_("Would you like us to protect these and try again?")):
					cache.clear()
					_upgrade(exclude, cache)
					sys.exit()
				sys.exit(
					_("{error} You have held broken packages").format(
						error=ERROR_PREFIX
					)
				)

			BrokenError(
				cache, tuple(pkg for pkg in cache if pkg.is_inst_broken)
			).broken_install()

		# Add any held pkgs to show in the table
		for pkg in is_upgrade:
			if (
				pkg.marked_upgrade
				or pkg.marked_delete
				or pkg in protected
				or not pkg.installed
			):
				continue

			if versions := pkg.versions:
				# After Upgrade they change the candidate for held packages.
				# This should get the latest version to show.
				cand = cast(Version, versions[0])
				nala_pkgs.held_pkgs.append(
					NalaPackage(
						pkg.name,
						cand.version,
						cand.installed_size,
						pkg.installed.version,
					)
				)
				continue
			# Should not hit this, but just in case
			nala_pkgs.held_pkgs.append(
				NalaPackage(pkg.name, "Unknown", 0, pkg.installed.version)
			)

		auto_remover(cache, nala_pkgs)
		get_changes(cache, nala_pkgs, "upgrade")

	_upgrade(exclude)


@nala.command(help=_("Install packages."))
def install(
	ctx: typer.Context,
	pkg_names: Optional[List[str]] = typer.Argument(
		None,
		metavar="PKGS ...",
		help=_("Package(s) to install"),
		autocompletion=package_completion,
	),
	purge: bool = PURGE,
	debug: bool = DEBUG,
	raw_dpkg: bool = RAW_DPKG,
	download_only: bool = DOWNLOAD_ONLY,
	remove_essential: bool = REMOVE_ESSENTIAL,
	update: bool = UPDATE,
	auto_remove: bool = AUTO_REMOVE,
	install_recommends: bool = RECOMMENDS,
	install_suggests: bool = SUGGESTS,
	fix_broken: bool = FIX_BROKEN,
	assume_yes: bool = ASSUME_YES,
	simple: bool = SIMPLE,
	dpkg_option: List[str] = OPTION,
	default_release: str = DEFAULT_RELEASE,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Install packages."""
	_install(pkg_names, ctx)


@nala.command(help=_("Remove packages."))
@nala.command("purge", help=_("Purge packages."))
@nala.command("uninstall", hidden=True)
def remove(
	pkg_names: List[str] = typer.Argument(
		...,
		metavar="PKGS ...",
		help=_("Package(s) to remove/purge"),
		autocompletion=remove_completion,
	),
	purge: bool = PURGE,
	debug: bool = DEBUG,
	raw_dpkg: bool = RAW_DPKG,
	download_only: bool = DOWNLOAD_ONLY,
	remove_essential: bool = REMOVE_ESSENTIAL,
	update: bool = UPDATE,
	auto_remove: bool = AUTO_REMOVE,
	fix_broken: bool = FIX_BROKEN,
	assume_yes: bool = ASSUME_YES,
	simple: bool = SIMPLE,
	dpkg_option: List[str] = OPTION,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Remove or Purge packages."""
	command_help("uninstall", "remove", update)
	_remove(pkg_names)


@nala.command("autoremove", help=_("Autoremove packages that are no longer needed."))
@nala.command("autopurge", help=_("Autopurge packages that are no longer needed."))
def _auto_remove(
	purge: bool = PURGE,
	debug: bool = DEBUG,
	config: bool = CONFIG,
	raw_dpkg: bool = RAW_DPKG,
	download_only: bool = DOWNLOAD_ONLY,
	remove_essential: bool = REMOVE_ESSENTIAL,
	update: bool = UPDATE,
	fix_broken: bool = FIX_BROKEN,
	assume_yes: bool = ASSUME_YES,
	simple: bool = SIMPLE,
	dpkg_option: List[str] = OPTION,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Command for autoremove."""
	sudo_check()
	if config and not arguments.is_purge():
		sys.exit(
			_(
				"{error} {config} must be used with either {autoremove} or {autopurge}."
			).format(
				error=ERROR_PREFIX,
				config=color("--config", "YELLOW"),
				autoremove=color("autoremove --purge", "YELLOW"),
				autopurge=color("autopurge", "YELLOW"),
			)
		)
	cache = setup_cache()
	check_state(cache, nala_pkgs)
	auto_remover(cache, nala_pkgs, config)
	get_changes(cache, nala_pkgs, "remove")


@nala.command(help=_("Show package details."))
@nala.command("info", hidden=True)
def show(
	pkg_names: List[str] = typer.Argument(
		...,
		help=_("Package(s) to show"),
		autocompletion=package_completion,
	),
	debug: bool = DEBUG,
	verbose: bool = VERBOSE,
	all_versions: bool = ALL_VERSIONS,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Show package details."""
	command_help("info", "show", None)
	cache = Cache()
	not_found: list[str] = []
	pkg_names = cache.glob_filter(pkg_names, show=True)
	pkg_names = cache.virtual_filter(pkg_names)
	additional_records = 0
	for num, pkg_name in enumerate(pkg_names):
		if pkg_name in cache:
			pkg = cache[pkg_name]
			additional_records += show_main(num, pkg)
			continue
		pkg_not_found(pkg_name, cache, not_found)

	if additional_records and not arguments.all_versions:
		additional_notice(additional_records)

	if not_found:
		for error in not_found:
			eprint(error)
		sys.exit(1)


@nala.command(help=_("Search package names and descriptions."))
def search(
	words: List[str] = typer.Argument(
		...,
		help=_("Regex or word to search for"),
		autocompletion=package_completion,
	),
	debug: bool = DEBUG,
	full: bool = FULL,
	names: bool = NAMES,
	installed: bool = INSTALLED,
	nala_installed: bool = NALA_INSTALLED,
	upgradable: bool = UPGRADABLE,
	upgradeable: bool = UPGRADEABLE,
	all_versions: bool = ALL_VERSIONS,
	all_arches: bool = ALL_ARCHES,
	virtual: bool = VIRTUAL,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Search package names and descriptions."""
	cache = Cache()
	user_installed = (
		get_list(get_history("Nala"), "User-Installed") if nala_installed else []
	)

	patterns = [
		compile_regex(fnmatch.translate(string[2:]))
		if string.startswith("g/")
		else compile_regex(string)
		for string in words
	]

	found: list[Package] = []
	arches = apt_pkg.get_architectures()
	for pkg in cache:
		if skip_pkg(cache, pkg, nala_installed, user_installed):
			continue
		if (
			pkg.architecture() in arches[0]
			or arguments.all_arches
			and pkg.architecture() in arches
		) and search_name(pkg, patterns):
			found.append(pkg)

	if not found:
		sys.exit(
			_("{error} {regex} not found.").format(
				error=ERROR_PREFIX, regex=", ".join(words)
			)
		)
	iter_search(found)


@nala.command("list", help=_("List packages based on package names."))
def list_pkgs(
	pkg_names: Optional[List[str]] = typer.Argument(
		None,
		help=_("Package(s) to list."),
		autocompletion=package_completion,
	),
	debug: bool = DEBUG,
	full: bool = FULL,
	installed: bool = INSTALLED,
	nala_installed: bool = NALA_INSTALLED,
	upgradable: bool = UPGRADABLE,
	upgradeable: bool = UPGRADEABLE,
	all_versions: bool = ALL_VERSIONS,
	virtual: bool = VIRTUAL,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""List packages based on package names."""
	cache = Cache()
	user_installed = (
		get_list(get_history("Nala"), "User-Installed") if nala_installed else []
	)

	patterns: list[Pattern[str]] = []
	if pkg_names:
		for name in pkg_names:
			if name.startswith("r/"):
				# Take out the prefix when we compile the regex
				patterns.append(compile_regex(name[2:]))
				continue
			# Otherwise we default to glob only for list
			patterns.append(compile_regex(fnmatch.translate(name)))

	def _list_gen() -> Generator[Package, None, None]:
		"""Generate to speed things up."""
		for pkg in cache:
			if skip_pkg(cache, pkg, nala_installed, user_installed):
				continue
			if pkg_names:
				for regex in patterns:
					# Match against the shortname so that arch isn't included
					if regex.match(pkg.shortname):
						yield pkg
						continue
				# If names were supplied and no matches
				# we don't want to grab everything
				continue

			# In this case no names were supplied so we list everything
			yield pkg

	if not iter_search(_list_gen()):
		sys.exit(_("Nothing was found to list."))


@nala.command(help=_("Clear out the local archive of downloaded package files."))
def clean(
	lists: bool = LISTS,
	fetch: bool = FETCH,
	debug: bool = DEBUG,
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
	color_force: bool = COLOR,
) -> None:
	"""Clear out the local archive of downloaded package files."""
	sudo_check()
	if lists:
		iter_remove(LISTS_DIR)
		iter_remove(LISTS_PARTIAL_DIR)
		print(_("Package lists have been cleaned"))
		return
	if fetch:
		NALA_SOURCES.unlink(missing_ok=True)
		print(_("Nala sources.list has been cleaned"))
		return
	iter_remove(ARCHIVE_DIR)
	iter_remove(PARTIAL_DIR)
	iter_remove(LISTS_PARTIAL_DIR)
	vprint(
		_("Removing {cache}\nRemoving {src_cache}").format(
			cache=PKGCACHE, src_cache=SRCPKGCACHE
		)
	)

	PKGCACHE.unlink(missing_ok=True)
	SRCPKGCACHE.unlink(missing_ok=True)
	print(_("Cache has been cleaned"))


@nala.command(hidden=True, help=_("I beg, pls moo"))
def moo(
	moos: Optional[List[str]] = typer.Argument(None, hidden=True),
	update: bool = typer.Option(None, hidden=True),
	color_force: bool = COLOR,
) -> None:
	"""I beg, pls moo."""
	print(CAT)
	can_no_moo = _("I can't moo for I'm a cat")
	print(f'{ELLIPSIS}"{can_no_moo}"{ELLIPSIS}')
	if update:
		what_did_you_expect = _("What did you expect to update?")
		print(f'{ELLIPSIS}"{what_did_you_expect}"{ELLIPSIS}')
		return
	if update is not None:
		what_did_you_expect = _("What did you expect no-update to do?")
		print(f'{ELLIPSIS}"{what_did_you_expect}"{ELLIPSIS}')



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Functions for Nala errors."""
from __future__ import annotations

import sys
from pathlib import Path
from typing import Iterable, NoReturn, Union

import apt_pkg
from apt.cache import FetchFailedException, LockFailedException
from apt.package import BaseDependency, Dependency, Package, Version

from nala import ROOT, _, color, color_version
from nala.cache import Cache
from nala.constants import ERROR_PREFIX, NOTICE_PREFIX, WARNING_PREFIX
from nala.debfile import NalaBaseDep, NalaDebPackage, NalaDep
from nala.dpkg import dpkg_error, update_error
from nala.rich import Columns, Text
from nala.search import BOT_LINE, LINE, TOP_LINE
from nala.show import format_dep
from nala.utils import dprint, eprint, term

DEPENDS = color(_("Depends:"))
"""'Depends:'"""
OR_DEPENDS = color(_("Either:"))
"""'Either:'"""
BREAKS = color(_("Breaks:"))
"""'Breaks:'"""
CONFLICTS = color(_("Conflicts:"))
"""'Conflicts:'"""
SECRET_VIRTUAL = _("{package} is only referenced by name, no packages provides it")
"""'{package} is a secret virtual package, nothing provides it'"""
BREAKS_MSG = _("{dependency} will break {package} {version}")
"""'{dependency} will break {package} {version}'"""
CONFLICTS_MSG = _("{dependency} conflicts with {package} {version}")
"""'{dependency} conflicts with {package} {version}'"""

NO_PROPER_ERR = _(
	"{error} python-apt gave us {apt_err} This isn't a proper error as it's empty"
)
AptErrorTypes = Union[
	FetchFailedException, LockFailedException, apt_pkg.Error, SystemError
]


class ExitCode:  # pylint: disable=too-few-public-methods
	"""Constants for Exit Codes."""

	SIGINT = 130
	SIGTERM = 143


class FileDownloadError(Exception):
	"""Exception class for passing errors.

	ERRHASH: 1 'Hash Sum is mismatched.'
	ENOENT: 2 'No such file or directory.'
	ERRSIZE: 3 'Size is mismatched.'
	"""

	ERRHASH = 1
	ENOENT = 2
	ERRSIZE = 3

	def __init__(  # pylint: disable=too-many-arguments
		self,
		error_str: str = "",
		errno: int = 0,
		filename: str = "",
		expected: str = "",
		received: str = "",
	) -> None:
		"""Define error properties."""
		super().__init__(error_str)
		self.error_str = error_str
		self.errno = errno
		self.filename = filename
		self.expected = expected
		self.received = received


class ParserError(Exception):
	"""Exception class for errors with parsing."""


# Should probably refactor this in the future. For now just disable the warning.
# pylint: disable=too-many-branches
def apt_error(apt_err: AptErrorTypes, update: bool = False) -> NoReturn | None:
	"""Take an error message from python-apt and formats it."""
	msg = f"{apt_err}"
	if not msg:
		if update_error:
			print()
			bad_mirror = False
			for line in update_error:
				bad_mirror = "Connection failed" in line
				eprint(line)
			if bad_mirror:
				eprint(
					_(
						"{notice} Some index files failed to download. "
						"They have been ignored, or old ones used instead."
					).format(notice=NOTICE_PREFIX)
				)

			if update:
				sys.exit(1)
		# Sometimes python apt gives us literally nothing to work with.
		# Probably an issue with sources.list. Needs further testing.
		if update:
			sys.exit(NO_PROPER_ERR.format(error=ERROR_PREFIX, apt_err=repr(apt_err)))
		return None

	if "installArchives() failed" in msg:
		eprint(_("{error} Installation has failed.").format(error=ERROR_PREFIX))
		eprint(
			_(
				"If you'd like to file a bug report please include '{debug_file}'"
			).format(debug_file=f"{ROOT}/var/log/nala/dpkg-debug.log")
		)
		sys.exit(1)

	if "," in msg:
		err_list = set(msg.split(","))
		for err in err_list:
			if "E:" in err:
				eprint(f"{ERROR_PREFIX} {err.replace('E:', '').strip()}")
				continue
			if "W:" in err:
				eprint(f"{WARNING_PREFIX} {err.replace('W:', '').strip()}")
				continue
		if update:
			sys.exit(1)
		return None

	eprint(f"{ERROR_PREFIX} {msg.replace('E:', '').strip()}")
	if not term.is_su():
		sys.exit(_("Are you root?"))
	if update or "Failed to lock directory" in msg:
		sys.exit(1)
	return None


def essential_error(pkg_list: list[Text]) -> NoReturn:
	"""Print error message for essential packages and exit."""
	print("=" * term.columns)
	print(_("{error} The following packages are essential!").format(error=ERROR_PREFIX))
	print("=" * term.columns)
	term.console.print(Columns(pkg_list, padding=(0, 2), equal=True))
	print("=" * term.columns)
	eprint(
		_("{error} You have attempted to remove essential packages").format(
			error=ERROR_PREFIX
		)
	)
	eprint(
		_("{error} Please use {switch} if you are sure you want to.").format(
			error=ERROR_PREFIX, switch=color("--remove-essential", "YELLOW")
		)
	)
	sys.exit(1)


def pkg_error(pkg_list: list[str], cache: Cache) -> NoReturn:
	"""Print error for package in list."""
	for pkg_name in pkg_list:
		if cache.is_any_virtual(pkg_name):
			eprint(
				_("{error} {package} has no installation candidate.").format(
					error=ERROR_PREFIX, package=color(pkg_name, "YELLOW")
				)
			)
			continue
		eprint(
			_("{error} {package} not found").format(
				error=ERROR_PREFIX, package=color(pkg_name, "YELLOW")
			)
		)
	sys.exit(1)


def print_dpkg_errors() -> None:
	"""Format and print dpkg errors if there are any."""
	if not dpkg_error:
		return
	# for line in dedupe_list(dpkg_error):
	for line in dpkg_error:
		if "dpkg:" in line:
			line = line.replace("dpkg:", "")
			if "warning:" in line:
				line = line.replace("warning:", "")
				if "downgrading" in line:
					line = line.replace("downgrading", "Downgraded")
				eprint(f"\n{WARNING_PREFIX} {line.strip()}")
				continue
			eprint(f"\n{ERROR_PREFIX} {line.strip()}")
			continue
		if "Errors were encountered" in line or "Processing was halted" in line:
			eprint(f"\n{line}")
			continue
		eprint(line)


def local_deb_error(error: apt_pkg.Error, name: str) -> NoReturn:
	"""Print what is wrong with the .deb and exit."""
	msg = f"{error}"
	if "Invalid archive signature" in msg:
		eprint(
			_("{error} {apt_error}\n  Unsupported File: {filename}").format(
				error=ERROR_PREFIX,
				apt_error=msg.replace("E:", "").strip(),
				filename=Path(name).resolve(),
			)
		)
		sys.exit(1)
	eprint(
		_("{error} {apt_error}\n  Could not read meta data from {filename}").format(
			error=ERROR_PREFIX, apt_error=msg, filename=Path(name).resolve()
		)
	)
	sys.exit(1)


class BrokenError:
	"""Calculate and print broken dependencies."""

	def __init__(
		self,
		cache: Cache,
		broken_list: Iterable[Package] | list[NalaDebPackage] | None = None,
	) -> None:
		"""Calculate and print broken install dependencies."""
		self.cache = cache
		self.broken_list = broken_list
		self.provides = (
			{
				ppkg
				for pkg in broken_list
				if isinstance(pkg, Package) and pkg.candidate
				for ppkg in pkg.candidate.provides
			}
			if broken_list
			else set()
		)

	def broken_install(self) -> int | NoReturn:
		"""Handle printing of errors due to broken packages."""
		# We have to clear the changes from the cache
		# before we can calculate why the packages are broken.
		if not self.broken_list:
			return 0
		self.cache.clear()
		if ret_count := sum(self.broken_pkg(pkg) for pkg in self.broken_list):
			self._print_held_error()
		else:
			return ret_count

	def broken_remove(self, broken_list: list[Package]) -> NoReturn:
		"""Calculate and print broken remove dependencies."""
		installed = tuple(
			pkg for pkg in self.cache if pkg.installed and pkg.installed.dependencies
		)
		self.cache.clear()
		for pkg in (
			pkg.name
			for pkg in broken_list
			if pkg.name in self._installed_dep_names(installed)
		):
			self._print_rdeps(pkg, installed)
		self._print_held_error()

	def held_pkgs(self, protected: set[Package]) -> None:
		"""Print packages that have been held back."""
		if not self.broken_list:
			# There is really not a chance of this happening
			return
		print(color(_("The following packages were kept back:"), "YELLOW"))
		if undetermined := [
			pkg for pkg in self.broken_list if not self.broken_pkg(pkg)
		]:
			print(
				color(
					_("The following were held due to exclusions:")
					if protected
					else _("Nala was unable to determine why these were held:")
				)
			)
			print(f"  {', '.join(color(pkg.name, 'YELLOW') for pkg in undetermined)}")

	def broken_pkg(self, pkg: Package | NalaDebPackage) -> int:
		"""Calculate and print broken Dependencies."""
		ret_count = 0
		version: NalaDebPackage | Version | None
		if isinstance(pkg, NalaDebPackage):
			version = pkg
		elif not (version := pkg.candidate):
			# We do this in case a broken package is locally installed.
			if not pkg.installed:
				return ret_count
			version = pkg.installed

		tree: list[str] = []
		dep_tree: list[str] = []
		arch = self._arch(pkg.name)

		indent = "    "
		if formatted_break := self.breaks_conflicts(pkg.name, version, arch):
			indent = LINE

		for dep in version.dependencies:
			dep_tree.extend(self._dep_tree(dep, arch, indent))

		if dep_tree:
			dep_tree.insert(0, f"{TOP_LINE if formatted_break else BOT_LINE} {DEPENDS}")
			dep_tree.append(dep_tree.pop().replace(TOP_LINE, BOT_LINE))
			tree.extend(dep_tree)

		tree.extend(f"{TOP_LINE} {_break}" for _break in formatted_break)
		if tree:
			tree.append(tree.pop().replace(TOP_LINE, BOT_LINE))
			tree.insert(0, color(pkg.name, "GREEN"))
			print("\n".join(tree), end="\n\n")
			ret_count += 1
		return ret_count

	def _dep_tree(
		self,
		dep: NalaDep | Dependency,
		arch: str,
		indent: str,
	) -> list[str]:
		dep_tree: list[str] = []
		if len(dep) > 1:
			count = 0
			or_tree = []
			for base_dep in dep:
				if formatted := self.format_broken(base_dep, arch):
					count += 1
					or_tree.append(formatted)
			if count == len(dep):
				start = f"{indent}{color(TOP_LINE, 'MAGENTA')}"
				for num, msg in enumerate(or_tree):
					if not num:
						dep_tree.append(f"{start} {msg}")
						continue
					dep_tree.append(f"{start} or {msg}")

		elif formatted := self.format_broken(dep[0], arch):
			dep_tree.append(f"{indent}{TOP_LINE} {formatted}")
		return dep_tree

	def format_broken(self, dep: BaseDependency | NalaBaseDep, arch: str = "") -> str:
		"""Format broken dependencies into a Tree, if any."""
		formatted_dep = format_dep(dep, 0).strip()
		dep_name = dep.name
		if arch and ":any" not in dep_name:
			dep_name = f"{dep_name}:{arch}"
			formatted_dep = formatted_dep.replace(dep.name, dep_name)
		# We print nothing on a virtual package
		if self.cache.is_virtual_package(dep_name):
			return ""
		if self.cache.is_secret_virtual(dep_name):
			return SECRET_VIRTUAL.format(package=formatted_dep)
		if dep_name not in self.cache:
			return _("{package} but it isn't in the cache").format(
				package=formatted_dep
			)
		# This means that the dependency doesn't have the right version in the cache
		if (
			dep.version
			and not dep.target_versions
			and not dep.installed_target_versions
		):
			dep_pkg = self.cache[dep.name]
			if (candidate := dep_pkg.candidate) and not apt_pkg.check_dep(
				candidate.version, dep.relation_deb, dep.version
			):
				return _("{package} but the cache version is {version}").format(
					package=formatted_dep,
					version=color_version(candidate.version),
				)
			# If none of our conditions are met we just fall back to a general error
			return _("{package} but it cannot be installed").format(
				package=formatted_dep
			)
		return ""

	def breaks_conflicts(
		self,
		pkg_name: str,
		version: Version | NalaDebPackage,
		arch: str,
	) -> tuple[str, ...]:
		"""Generate tree objects for breaks and conflict type deps."""
		break_conflict: list[str] = []
		for dep_type in ("Breaks", "Conflicts"):
			if deps := version.get_dependencies(dep_type):
				dprint(f"{pkg_name} {dep_type}:\n{deps}")
				break_conflict.extend(
					self.format_broken_conflict(
						deps,
						BREAKS_MSG if dep_type == "Breaks" else CONFLICTS_MSG,
						arch,
					)
				)
		return tuple(break_conflict)

	def format_broken_conflict(
		self,
		breaks: list[Dependency] | list[NalaDep],
		dep_string: str,
		arch: str = "",
	) -> list[str]:
		"""Format broken conflict/breaks dependency into a Tree."""
		break_tree: list[str] = []
		for dep in breaks:
			if not (target_versions := dep.target_versions):
				continue
			if installed_versions := tuple(
				ver for ver in target_versions if ver.is_installed
			):
				break_tree.append(
					dep_string.format(
						dependency=self._dependency_name(dep, arch),
						package=color(
							self._break_pkg_name(installed_versions[0], arch), "GREEN"
						),
						version=color_version(installed_versions[0].version),
					)
				)

			if not (target_provides := target_versions[0].provides):
				continue

			if conflict := self.format_conflict(dep, target_provides, dep_string):
				break_tree.append(conflict)
		return break_tree

	def format_conflict(
		self,
		dep: Dependency | NalaDep,
		target_provides: list[str],
		dep_string: str,
	) -> str | None:
		"""Format a conflicting package that isn't installed."""
		return (
			next(
				(
					dep_string.format(
						dependency=format_dep(dep[0]).strip(),
						package=color(provide_name, "GREEN"),
						version="",
					)
					for provide_name in target_provides
					if dep[0].name != provide_name and provide_name in self.provides
				),
				None,
			)
			if self.provides
			else None
		)

	@staticmethod
	def _break_pkg_name(version: Version, arch: str) -> str:
		"""Get the name of the package that is broken."""
		return (
			f"{version.package.name}:{version.architecture}"
			if arch
			else version.package.name
		)

	@staticmethod
	def _dependency_name(dep: Dependency | NalaDep, arch: str) -> str:
		"""Get the formatted dependency name."""
		return (
			format_dep(dep[0]).replace(dep[0].name, f"{dep[0].name}:{arch}")
			if arch
			else format_dep(dep[0])
		).strip()

	@staticmethod
	def _arch(pkg_name: str) -> str:
		return (
			pkg_name.split(":")[1]
			if ":" in pkg_name
			and all(substring not in pkg_name for substring in (":all", ":any"))
			else ""
		)

	@staticmethod
	def unmarked_error(pkgs: list[Package]) -> None:
		"""Print error messages related to the fixer unmarking packages requested for install."""
		for pkg in pkgs:
			if not pkg.marked_upgrade or pkg.marked_install:
				print(
					_("{package} has been unmarked.").format(
						package=color(pkg.name, "GREEN"),
					)
				)
		print(
			_("Try {switch} if you're sure they can be installed.").format(
				switch=color("--no-fix-broken", "YELLOW")
			)
		)
		sys.exit(
			_("{error} Some packages were unable to be installed.").format(
				error=ERROR_PREFIX
			)
		)

	@staticmethod
	def _installed_dep_names(installed_pkgs: tuple[Package, ...]) -> tuple[str, ...]:
		"""Iterate installed pkgs and return all of their deps in a tuple.

		This is so we can reduce iterations when checking reverse depends.
		"""
		total_deps = set()
		for pkg in installed_pkgs:
			if not (pkg_installed := pkg.installed):
				continue
			for deps in pkg_installed.dependencies:
				for dep in deps:
					total_deps.add(dep.name)
		return tuple(total_deps)

	@staticmethod
	def _print_rdeps(name: str, installed_pkgs: tuple[Package, ...]) -> int:
		"""Print the installed reverse depends of a package."""
		msg = color(
			_("Installed packages that depend on {package}").format(
				package=color(name, "GREEN")
			)
			+ "\n",
			"YELLOW",
		)
		for pkg in installed_pkgs:
			if not (pkg_installed := pkg.installed):
				continue
			for dep in pkg_installed.dependencies:
				if name in dep.rawstr:
					dep_msg = f"  {color(pkg.name, 'GREEN')}"
					if pkg.essential:
						dep_msg = _("{package} is an Essential package!").format(
							package=dep_msg
						)
					msg += f"{dep_msg}\n"
					break
		print(msg.strip())
		return 1

	@staticmethod
	def _print_held_error() -> NoReturn:
		"""Print the held broken error and exit."""
		eprint(
			_("{notice} The information above may be able to help").format(
				notice=NOTICE_PREFIX
			)
		)
		sys.exit(_("{error} You have held broken packages").format(error=ERROR_PREFIX))



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Module for file constants."""
from __future__ import annotations

import re
from enum import IntEnum
from pathlib import Path
from signal import Handlers  # pylint: disable=no-name-in-module #Codacy
from types import FrameType
from typing import Any, Callable, Optional, Union

import apt_pkg

from nala import ROOT, USR, _, color

# File Constants
GPL3_LICENSE = Path(f"{USR}/share/common-licenses/GPL-3")
"""/usr/share/common-licenses/GPL-3"""
OLD_NALA_SOURCES = Path(f"{ROOT}/etc/apt/sources.list.d/nala-sources.list")
"""/etc/apt/sources.list.d/nala-sources.list"""
NALA_SOURCES = Path(f"{ROOT}/etc/apt/sources.list.d/fetch.sources")
"""/etc/apt/sources.list.d/fetch.sources"""
NALA_DIR = Path(f"{ROOT}/var/lib/nala")
"""/var/lib/nala"""
NALA_LOGDIR = Path(f"{ROOT}/var/log/nala")
"""/var/log/nala"""
NALA_DEBUGLOG = NALA_LOGDIR / "nala-debug.log"
"""/var/log/nala/nala.debug.log"""
DPKG_LOG = NALA_LOGDIR / "dpkg-debug.log"
"""/var/log/nala/dpkg-debug.log"""
DPKG_STATUS_LOG = NALA_LOGDIR / "dpkg-status.log"
"""/var/log/nala/dpkg-status.log"""
NALA_TERM_LOG = Path(f"{ROOT}/var/log/nala/term.log")
"""/var/log/nala/term.log"""
NALA_HISTORY = Path(f"{ROOT}/var/lib/nala/history.json")
"""/var/lib/nala/history.json"""
PACSTALL_METADATA = Path(f"{ROOT}/var/log/pacstall/metadata")
"""/var/log/pacstall/metadata"""
REBOOT_REQUIRED = Path(f"{ROOT}/var/run/reboot-required")
"""/var/run/reboot-required"""
REBOOT_PKGS = Path(f"{ROOT}/var/run/reboot-required.pkgs")
"""'/var/run/reboot-required.pkgs'"""
NALA_LOCK_FILE = Path(f"{ROOT}/var/lock/nala.lock")
"""/var/lock/nala.lock"""

CACHE_DIR = Path(apt_pkg.config.find_dir("Dir::Cache", "/var/cache/apt"))
ETC_DIR = Path(apt_pkg.config.find_dir("Dir::Etc", "/etc/apt"))


def apt_file(key: str, prepend: Path, sub_path: str) -> Path:
	"""Get the apt directory. Build prepend/sub_path if None."""
	if file := apt_pkg.config.find_file(key):
		return Path(file)
	return prepend / sub_path


# Apt Directories
ARCHIVE_DIR = apt_file("Dir::Cache::Archives", CACHE_DIR, "/var/cache/apt/archives")
PARTIAL_DIR = ARCHIVE_DIR / "partial"
PKGCACHE = apt_file("Dir::Cache::pkgcache", CACHE_DIR, "pkgcache.bin")
SRCPKGCACHE = apt_file("Dir::Cache::srcpkgcache", CACHE_DIR, "srcpkgcache.bin")
LISTS_DIR = Path(apt_pkg.config.find_dir("Dir::State::Lists", "/var/lib/apt/lists/"))
LISTS_PARTIAL_DIR = LISTS_DIR / "partial"
DPKG_STATE = Path(apt_pkg.config.find_dir("Dir::State::status", "/var/lib/dpkg/status"))
SOURCELIST = apt_file("Dir::Etc::sourcelist", ETC_DIR, "/etc/apt/sources.list")
SOURCEPARTS = apt_file("Dir::Etc::sourceparts", ETC_DIR, "/etc/apt/sources.list.d")

# pylint: disable=invalid-name
HANDLER = Union[Callable[[int, Optional[FrameType]], Any], int, Handlers, None]

ERROR_PREFIX = color(_("Error:"), "RED")
WARNING_PREFIX = color(_("Warning:"), "YELLOW")
NOTICE_PREFIX = color(_("Notice:"), "YELLOW")

# Compiled Regex
ERRNO_PATTERN = re.compile(r"\[.*\]")


class SelectedState(IntEnum):
	"""Selected state constants for pkgs."""

	UNKNOWN = 0
	INSTALL = 1
	HOLD = 2
	DEINSTALL = 3
	PURGE = 4


class CurrentState(IntEnum):
	"""Current state constants for pkgs."""

	NOT_INSTALLED = 0
	UNPACKED = 1
	HALF_CONFIGURED = 2
	HALF_INSTALLED = 4
	CONFIG_FILES = 5
	INSTALLED = 6
	TRIGGERS_AWAITED = 7
	TRIGGERS_PENDING = 8


class InstState(IntEnum):
	"""Installed state constants for pkgs."""

	OK = 0
	REINSTREQ = 1
	HOLD_INST = 2
	HOLD_REINSTREQ = 3


# dpkg constants
NOTICES = (
	b"The currently running kernel version is not the expected kernel version",
	b"Please remove.",
	b"NOTICE:",
	b"Warning:",
)
SPAM = (
	# Stuff that's pretty useless
	"(Reading database",  #'(Reading database ... 247588 files and directories currently installed.)'
	"files and directories currently installed.)",
	"Selecting previously unselected package",  # 'Selecting previously unselected package chafa.'
	"Preparing to unpack",  # 'Preparing to unpack .../2-chafa_1.8.0-1_amd64.deb ...'
	"Extracting templates from packages:",
	"Fetched",
)
DPKG_STATUS = (
	b"Scanning processes...",
	b"Scanning candidates...",
	b"Scanning linux images...",
	b"Extracting templates from packages",
	b"Reading changelogs...",
	b"Retrieving bug reports...",
	b"Parsing Found/Fixed information...",
	b"[Working]",
	b"[Connecting",
	b"[Waiting for headers]",
	b"[Connected to",
)
DPKG_ERRORS = (
	b"dpkg:",
	b"failed to allocate memory for new entry",
	b"error processing package",
	b"error processing archive",
	b"Errors were encountered while processing",
	b"Processing was halted because there were too many errors",
)
CAT = r"""
   |\---/|
   | ,_, |
    \_`_/-..----.
 ___/ `   ' ,""+ \  sk
(__...'   __\    |`.___.';
  (_,...'(_,.`__)/'.....+
"""



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2010 - 2021 Tatsuhiro Tsujikawa
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
# nala is based upon apt-metalink https://github.com/tatsuhiro-t/apt-metalink
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Where Utilities who don't have a special home come together."""
from __future__ import annotations

import contextlib
import os
import re
import signal
import sys
import termios
import tty
from dataclasses import dataclass, field
from datetime import datetime
from fcntl import LOCK_EX, LOCK_NB, lockf
from pathlib import Path
from types import FrameType
from typing import TYPE_CHECKING, Any, Generator, Iterable, Pattern

from apt.package import Package, Version

from nala import TERMUX, _, color, console
from nala.constants import (
	ERROR_PREFIX,
	HANDLER,
	NALA_DEBUGLOG,
	NALA_DIR,
	NALA_LOCK_FILE,
	NALA_LOGDIR,
	NOTICE_PREFIX,
	PARTIAL_DIR,
)
from nala.options import arguments
from nala.rich import from_ansi

if TYPE_CHECKING:
	from nala.debfile import NalaDebPackage
	from nala.fetch import FetchLive

LOCK_FILE = None

# NOTE: Answers for the Question prompt "[Y/n]"
YES_NO = _("Y/y N/n").split()
YES = YES_NO[0].split("/")
NO = YES_NO[1].split("/")


class Terminal:
	"""Represent the user terminal."""

	# Term Constants
	STDIN = 0
	STDOUT = 1
	STDERR = 2

	# Control Codes
	CURSER_UP = b"\x1b[1A"
	CURSER_DOWN = b"\x1b[1B"
	CURSER_FORWARD = b"\x1b[1C"
	CURSER_BACK = b"\x1b[1D"
	CLEAR_LINE = b"\x1b[2k"
	CLEAR = b"\x1b[2J"
	CLEAR_FROM_CURRENT_TO_END = b"\x1b[K"
	BACKSPACE = b"\x08"
	HOME = b"\x1b[H"
	ENABLE_BRACKETED_PASTE = b"\x1b[?2004h"
	DISABLE_BRACKETED_PASTE = b"\x1b[?2004l"
	ENABLE_ALT_SCREEN = b"\x1b[?1049h"
	DISABLE_ALT_SCREEN = b"\x1b[?1049l"
	SHOW_CURSOR = b"\x1b[?25h"
	HIDE_CURSOR = b"\x1b[?25l"
	SET_CURSER = b"\x1b[?1l"
	SAVE_TERM = b"\x1b[22;0;0t"
	RESTORE_TERM = b"\x1b[23;0;0t"
	APPLICATION_KEYPAD = b"\x1b="
	NORMAL_KEYPAD = b"\x1b>"
	CR = b"\r"
	LF = b"\n"
	CRLF = b"\r\n"

	def __init__(self) -> None:
		"""Represent the user terminal."""
		self.console = console
		self.mode: list[int | list[bytes | int]] = []
		self.term_type: str = os.environ.get("TERM", "").lower()
		self.locale: str = ""
		self.set_environment()

	def __repr__(self) -> str:
		"""Represent state of the user terminal as a string."""
		kwarg = "\n    ".join(
			(f"{key} = {value},") for key, value in self.__dict__.items()
		)
		return f"Terminal = [\n    {kwarg}\n]"

	def set_environment(self) -> None:
		"""Check and set various environment variables."""
		# Termios can't run if we aren't in a terminal
		# Just catch the exception and continue.
		if self.can_format():
			self.mode = termios.tcgetattr(self.STDIN)

		if self.lines < 13 or self.columns < 31:
			print(
				_("Terminal can't support dialog, falling back to readline"),
				file=sys.stderr,
			)
			os.environ["DEBIAN_FRONTEND"] = "readline"
		# Readline is too hard to support with our fancy formatting
		if os.environ.get("DEBIAN_FRONTEND") == "readline":
			arguments.raw_dpkg = True
		os.environ["DPKG_COLORS"] = "never"
		self.locale = os.environ.get("LANG", "")
		# We have to set lang as C so we get predictable output from dpkg.
		os.environ["LANG"] = "C" if self.console.options.ascii_only else "C.UTF-8"

	@property
	def columns(self) -> int:
		"""Return termindal width."""
		return self.console.width

	@property
	def lines(self) -> int:
		"""Return terminal height."""
		return self.console.height

	def can_format(self) -> bool:
		"""Return if we're allowed to do anything fancy."""
		return (
			os.isatty(self.STDOUT)
			and os.isatty(self.STDIN)
			and self.term_type not in ("dumb", "unknown")
		)

	def restore_mode(self) -> None:
		"""Restore the mode the Terminal was initialized with."""
		if self.can_format():
			termios.tcsetattr(self.STDIN, termios.TCSAFLUSH, self.mode)

	def restore_locale(self) -> None:
		"""Restore the locale to it's original value."""
		os.environ["LANG"] = self.locale

	def set_raw(self) -> None:
		"""Set terminal raw."""
		if self.can_format():
			tty.setraw(self.STDIN)

	def write(self, data: bytes) -> None:
		"""Write bytes directly to stdout."""
		os.write(self.STDOUT, data)

	def is_xterm(self) -> bool:
		"""Return True if we're in an xterm, False otherwise."""
		return "xterm" in self.term_type

	@staticmethod
	def is_su() -> bool:
		"""Return True if we're super user and False if we're not."""
		return TERMUX or os.geteuid() == 0


class DelayedKeyboardInterrupt:
	"""Context manager to delay KeyboardInterrupt.

	Keyboard Interrupts will be delayed until out of scope.
	"""

	def __init__(self) -> None:
		"""Context manager to delay KeyboardInterrupt."""
		self.signal_received: tuple[int, FrameType | None] | bool
		self.old_handler: HANDLER

	def __enter__(self) -> None:
		"""Enter context."""
		self.signal_received = False
		self.old_handler = signal.signal(signal.SIGINT, self.handler)

	def handler(self, sig: int, frame: FrameType | None) -> None:
		"""Handle sigint signals."""
		self.signal_received = (sig, frame)
		dprint("SIGINT received. Delaying KeyboardInterrupt.")

	def __exit__(self, _type: None, _value: None, _traceback: None) -> None:
		"""Exit context."""
		signal.signal(signal.SIGINT, self.old_handler)
		if isinstance(self.signal_received, tuple) and callable(self.old_handler):
			self.old_handler(*self.signal_received)


@dataclass
class PackageHandler:  # pylint: disable=too-many-instance-attributes
	"""Class for storing package lists."""

	autoremoved: set[str] = field(default_factory=set)
	user_explicit: list[Package] = field(default_factory=list)
	local_debs: list[NalaDebPackage] = field(default_factory=list)
	# Packages that can be auto removed, but won't
	not_needed: list[NalaPackage] = field(default_factory=list)
	delete_pkgs: list[NalaPackage] = field(default_factory=list)
	install_pkgs: list[NalaPackage] = field(default_factory=list)
	reinstall_pkgs: list[NalaPackage] = field(default_factory=list)
	upgrade_pkgs: list[NalaPackage] = field(default_factory=list)
	autoremove_pkgs: list[NalaPackage] = field(default_factory=list)
	autoremove_config: list[NalaPackage] = field(default_factory=list)
	delete_config: list[NalaPackage] = field(default_factory=list)
	recommend_pkgs: list[NalaPackage | list[NalaPackage]] = field(default_factory=list)
	suggest_pkgs: list[NalaPackage | list[NalaPackage]] = field(default_factory=list)
	configure_pkgs: list[NalaPackage] = field(default_factory=list)
	downgrade_pkgs: list[NalaPackage] = field(default_factory=list)
	held_pkgs: list[NalaPackage] = field(default_factory=list)

	def no_summary(
		self, pkg_set: list[NalaPackage] | list[NalaPackage | list[NalaPackage]]
	) -> bool:
		"""Return True if we shouldn't print a summary for the package set."""
		return pkg_set in (self.suggest_pkgs, self.recommend_pkgs, self.not_needed)

	def all_pkgs(self) -> Generator[NalaPackage | NalaDebPackage, None, None]:
		"""Return a list of all the packages to be altered."""
		yield from (
			self.delete_pkgs
			+ self.autoremove_pkgs
			+ self.install_pkgs
			+ self.reinstall_pkgs
			+ self.downgrade_pkgs
			+ self.upgrade_pkgs
			+ self.configure_pkgs
			+ self.autoremove_config
			+ self.delete_config
		)
		yield from self.local_debs

	def dpkg_progress_total(self) -> int:
		"""Calculate our total operations for the dpkg progress bar."""
		return (
			len(
				self.delete_pkgs
				+ self.autoremove_pkgs
				+ self.install_pkgs
				+ self.reinstall_pkgs
				+ self.downgrade_pkgs
				+ self.upgrade_pkgs
				+ self.configure_pkgs
			)
			* 2
			# For local deb installs we add 1 more because of having to start
			# and stop InstallProgress an extra time for each package
			+ len(self.local_debs)
			# Purging configuration files only have 1 message
			+ len(self.autoremove_config + self.delete_config)
			# This last +1 for the ending of dpkg itself
			+ 1
		)


@dataclass
class NalaPackage:
	"""Class that represents a Nala package."""

	name: str
	version: str
	size: int
	old_version: str | None = None

	@property
	def unit_size(self) -> str:
		"""Return the size as a readable unit. Example 12 MB."""
		return unit_str(self.size)


term = Terminal()


def command_help(wrong: str, correct: str, update: bool | None) -> None:
	"""Check if the user typed a common mistake of a command."""
	command_ask = False
	if arguments.command == "history" and arguments.history == wrong:
		arguments.command = correct
		wrong = f"history {wrong}"
		command_ask = True

	elif arguments.command == wrong:
		arguments.command = correct
		command_ask = True

	if command_ask:
		arguments.set_update(update)
		if not ask(
			_("{command} is not a command\nDid you mean {correction}?").format(
				command=color(wrong, "YELLOW"),
				correction=color(f"nala {correct}", "YELLOW"),
			)
		):
			sys.exit(1)


def ask(question: str) -> bool:
	"""Ask the user {question}.

	resp = input(f'{question}? [Y/n]

	Y returns True
	N returns False
	"""
	with contextlib.suppress(AttributeError):
		if arguments.assume_yes:
			return True
		if arguments.assume_no:
			return False

	resp = input(f"{question} [{YES[0]}/{NO[1]}] ").strip()
	return len(resp) == 0 or resp[0] in (YES[0], YES[1], "Y", "y")


def unauth_ask(question: str) -> bool:
	"""Ask the user if they'd like to accept unauthenticated packages."""
	if not arguments.config.apt.find_b("APT::Get::AllowUnauthenticated", False):
		# If a user is piping something into Nala to bypass this prompt, error because this is unsafe.
		# The option should be passed on the command line so it's explicit what is happening in scripts.
		if arguments.assume_yes or not sys.stdin.isatty():
			sys.exit(
				_(
					"{error} Some packages are unable to be authenticated. Use "
					"'-o APT::Get::AllowUnauthenticated=true' with `--yes`"
				).format(error=ERROR_PREFIX)
			)
		return ask(f"{NOTICE_PREFIX} {question}")
	return True


def compile_regex(regex: str) -> Pattern[str]:
	"""Compile regex and exit on failure."""
	try:
		return re.compile(regex, re.IGNORECASE)
	except re.error as error:
		sys.exit(
			_(
				"{error} failed regex compilation '{error_msg} at position {position}'"
			).format(error=ERROR_PREFIX, error_msg=error.msg, position=error.pos)
		)


def sudo_check(args: Iterable[str] | None = None) -> None:
	"""Check for root and exit if not root."""
	if not term.is_su():
		if arguments.command == "install" and arguments.fix_broken and not args:
			sys.exit(
				_("{error} Nala needs root to fix broken packages").format(
					error=ERROR_PREFIX
				)
			)
		sys.exit(
			_("{error} Nala needs root to {command}").format(
				error=ERROR_PREFIX, command=arguments.command
			)
		)
	# Make sure our directories exist
	NALA_DIR.mkdir(exist_ok=True)
	NALA_LOGDIR.mkdir(exist_ok=True)
	PARTIAL_DIR.mkdir(parents=True, exist_ok=True)

	NALA_LOCK_FILE.touch(exist_ok=True)
	global LOCK_FILE  # pylint: disable=global-statement
	# pylint: disable=consider-using-with
	LOCK_FILE = NALA_LOCK_FILE.open("r+", encoding="ascii")
	current_pid = os.getpid()
	last_pid = LOCK_FILE.read()

	try:
		dprint("Setting Lock")
		lockf(LOCK_FILE, LOCK_EX | LOCK_NB)
		LOCK_FILE.seek(0)
		LOCK_FILE.write(f"{current_pid}")
		LOCK_FILE.truncate()
		dprint("Lock Set")
	except OSError:
		sys.exit(
			_("{error} Nala is already running another instance {last_pid}").format(
				error=ERROR_PREFIX, last_pid=color(last_pid, "YELLOW")
			)
		)


def get_date() -> str:
	"""Return the formatted Date and Time."""
	return f"{datetime.now().astimezone().strftime('%Y-%m-%d %H:%M:%S %Z')}"


def unit_str(val: int) -> str:
	"""Check integer and figure out what format it should be.

	`unit_str` will return a string with a leading space like " 12 MB".

	You need to strip `unit_str` if you do not want the space.
	"""
	if arguments.config.get_bool("filesize_binary", False):
		base = 1024
		size = ("Bytes", "KiB", "MiB", "GiB", "TiB", "PiB", "EiB", "ZiB", "YiB")
	else:
		base = 1000
		size = ("Bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")

	if val > base**3:
		return f"{val/base**3 :.1f} {size[3]}"
	if val > base**2:
		return f"{val/base**2 :.1f} {size[2]}"
	if val > base:
		return f"{round(val/1000) :.0f} {size[1]}"
	return f"{val :.0f} {size[0]}"


def iter_remove(path: Path) -> None:
	"""Iterate the directory supplied and remove all files."""
	vprint(_("Removing files in {dir}").format(dir=path))
	for file in path.iterdir():
		if file.is_file():
			vprint(_("Removed: {filename}").format(filename=file))
			file.unlink(missing_ok=True)


def get_version(
	pkg: Package, cand_first: bool = False, inst_first: bool = False
) -> Version | tuple[Version, ...]:
	"""Get the version, any version of a package."""
	if not cand_first and arguments.all_versions:
		return tuple(pkg.versions)
	if cand_first:
		return pkg.candidate or pkg.installed or pkg.versions[0]
	if inst_first:
		return pkg.installed or pkg.candidate or pkg.versions[0]
	for version in pkg.versions:
		return version
	# It would be really weird if we ever actually hit this error
	sys.exit(
		_("{error} can't find version for {package}").format(
			error=ERROR_PREFIX, package=pkg.name
		)
	)


def get_pkg_version(
	pkg: Package, cand_first: bool = False, inst_first: bool = False
) -> Version:
	"""Get the version."""
	if cand_first:
		return pkg.candidate or pkg.installed or pkg.versions[0]
	if inst_first:
		return pkg.installed or pkg.candidate or pkg.versions[0]
	for version in pkg.versions:
		return version
	# It would be really weird if we ever actually hit this error
	sys.exit(
		_("{error} can't find version for {package}").format(
			error=ERROR_PREFIX, package=pkg.name
		)
	)


def get_pkg_name(candidate: Version) -> str:
	"""Return the package name.

	Checks if we need and epoch in the path.
	"""
	if ":" in candidate.version:
		index = candidate.version.index(":")
		epoch = f"_{candidate.version[:index]}%3a"
		return Path(candidate.filename).name.replace("_", epoch, 1)
	return Path(candidate.filename).name


def pkg_candidate(pkg: Package) -> Version:
	"""Type enforce package candidate."""
	assert pkg.candidate
	return pkg.candidate


def pkg_installed(pkg: Package) -> Version:
	"""Type enforce package installed."""
	assert pkg.installed
	return pkg.installed


def dedupe_list(original: Iterable[str]) -> list[str]:
	"""Deduplicate a list.

	Useful for when we want to maintain the list order and can't use set()
	"""
	dedupe = []
	for item in original:
		if item not in dedupe:
			dedupe.append(item)
	return dedupe


def vprint(msg: object) -> None:
	"""Print message if verbose."""
	if arguments.verbose or arguments.debug:
		print(msg)
	if arguments.debug:
		dprint(from_ansi(f"{msg}").plain, from_verbose=True)
	if (stdout := sys.stdout) is not None:
		stdout.flush()


def dprint(msg: object, from_verbose: bool = False) -> None:
	"""Print message if debugging, write to log if root.

	from_verbose as true will stop this from printing.

	vprint sends it's messages here to be put in the log.
	"""
	if not arguments.debug:
		return
	if not from_verbose:
		print(f"DEBUG: {msg}")
	if term.is_su():
		with open(NALA_DEBUGLOG, "a", encoding="utf-8") as logfile:
			logfile.write(f"[{get_date()}] DEBUG: {msg}\n")


def eprint(*args: Any, **kwargs: Any) -> None:
	"""Print message to stderr."""
	print(*args, file=sys.stderr, **kwargs)



#!/usr/bin/python3
"""Management tools related to building Nala."""
import os
import sys
import time
from pathlib import Path
from subprocess import run

import typer

from nala import USR, __version__ as version

PO_FILES = tuple(Path("po").glob("*.po"))
SOURCE_FILES = tuple(Path("nala").glob("*.py"))
DOCS_DIR = Path("docs")

# pylint: disable=too-few-public-methods
class BuildEnvironment:
	"""Hold environment variables."""

	def __init__(self, build_dir: str) -> None:
		"""Hold environment variables."""
		self.build_dir = build_dir
		self.bin_dir = f"{build_dir}{USR}/bin"
		self.locale_dir = f"{build_dir}{USR}/share/locale"


nala_app = typer.Typer(add_completion=False)


def check_root(operation: str) -> None:
	"""Check for root and exit if not."""
	if os.getuid() != 0:
		sys.exit(f"Error: You need root to install the {operation}.")


@nala_app.command(name="man")
def convert_man(
	install: bool = typer.Option(
		False, "--install", help="Additionally install the man pages"
	)
) -> None:
	"""Convert .rst files into man pages."""
	if install:
		check_root("man pages")

	date = time.strftime(
		"%d %b %Y",
		time.gmtime(int(os.environ.get("SOURCE_DATE_EPOCH", time.time()))),
	)

	# Convert man page and install if requested
	for file in DOCS_DIR.iterdir():
		if not file.name.endswith(".rst"):
			continue

		# If the install switch is set then we install it to the man directory.
		man_page = Path(
			f"{USR}/share/man/man8/{file.name.replace('.rst', '')}"
			if install
			else f"{file}".replace(".rst", "")
		)

		print(f"Installing {file} -> {man_page}")

		pandoc = [
			"pandoc",
			f"{file}",
			f"--output={man_page}",
			"--standalone",
			"--variable=header:'Nala User Manual'",
			f"--variable=footer:{version}",
			f"--variable=date:{date}",
			"--variable=section:8",
			"--from",
			"rst",
			"--to",
			"man",
		]
		run(pandoc, check=True)


def update_translations() -> None:
	"""Update the .po files from the pot file."""
	update = "pybabel update --no-wrap -i po/nala.pot".split()
	for path in PO_FILES:
		# Strip off `.po`
		run(update + ["-o", f"po/{path.name}", "-l", path.name[:-3]], check=True)


def compile_translations(env: BuildEnvironment) -> None:
	"""Compile .po files to .mo."""
	pybable = f"pybabel compile --directory={env.locale_dir} --domain=nala --use-fuzzy".split()
	for path in PO_FILES:
		# Strip off `.po`
		locale = path.name[:-3]
		Path(f"{env.locale_dir}/{locale}/LC_MESSAGES/").mkdir(
			parents=True, exist_ok=True
		)

		compile_mo = pybable + [f"--input-file=po/{locale}.po", f"--locale={locale}"]
		run(compile_mo, check=True)


def extract_translations() -> None:
	"""Extract translations to nala.pot."""
	run(
		[
			"pybabel",
			"extract",
			"--add-comments=NOTE:",
			"--strip-comments",
			"--project=nala",
			"--copyright-holder=Blake Lee",
			f"--version={version}",
			"--msgid-bugs-address=https://gitlab.com/volian/nala/-/issues",
			"--no-wrap",
			*SOURCE_FILES,
			"-o",
			"po/nala.pot",
		],
		check=True,
	)


@nala_app.command()
def babel(
	extract: bool = typer.Option(
		False,
		"--extract",
		help="Extract translations to nala.pot and update the po files.",
	),
	_compile: bool = typer.Option(False, "--compile", help="Compile .po files to .mo"),
	install: bool = typer.Option(
		False, "--install", help="Additionally install the translation files."
	),
) -> None:
	"""Manage translation files."""
	if extract:
		extract_translations()
		update_translations()
	elif _compile:
		if install:
			check_root("translation files")

		compile_translations(
			BuildEnvironment(build_dir="")
			if install
			else BuildEnvironment(build_dir="debian/nala"),
		)
	else:
		sys.exit("Error: You need to specify either '--compile' or '--extract'")


if __name__ == "__main__":
	nala_app()



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Nala package manager."""

from __future__ import annotations

import os
import sys
from subprocess import run

# pylint: disable=subprocess-run-check

# Stop this code from running during shell completions
if not os.environ.get("_NALA_COMPLETE"):
	# If nala is ran with no arguments so the man page
	if len(sys.argv) == 1:
		sys.exit(run(["man", "nala"]).returncode)


# Set Path as below for termux environment
# ROOT = "@TERMUX_PREFIX@"
# USR = "@TERMUX_PREFIX@"
# TERMUX = True
ROOT = ""
USR = "/usr"
TERMUX = False

# pylint: disable=wrong-import-position
# Before we import anything else we need to make sure that this is in our path
# Else apt_pkg will give an import error. Conda is an example of this breaking.
# Additionally if something is installed with pip and pulls in dependencies
# That Nala uses this can also break. This fix is to add dist-packages to the beginning
# So it's always queried first.
sys.path.insert(0, f"{USR}/lib/python3/dist-packages")

import gettext

import apt_pkg
from rich.console import Console

# Debian needs to be initialized before the apt config.
# Inside debian.debian_support they init apt_pkg without checking if it's been initialized.
# If this is imported after apt_pkg.config is setup,
# it causes duplicates for things like pre/post invoke hooks
# https://gitlab.com/volian/nala/-/issues/287
import debian.debian_support  # isort:skip

if "APT" not in apt_pkg.config:
	apt_pkg.init_config()
apt_pkg.init_system()

__version__ = "0.16.0"

console = Console()

COLOR_CODES: dict[str, str | int] = {
	"RESET": "\x1b[0m",
	"ITALIC": "\x1b[3m",
	"RED": 31,
	"GREEN": 32,
	#'YELLOW' : 33,
	# We set Yellow as more of an orange for light terminals.
	"YELLOW": "38;5;202",
	"BLUE": 34,
	"MAGENTA": 35,
	"CYAN": 36,
	"WHITE": 37,
}


def color(text: object, text_color: str = "") -> str:
	"""Return colored text if allowed."""
	if os.environ.get("FORCE_COLOR"):
		return color_text(text, text_color)
	if (
		os.environ.get("NO_COLOR")
		or not console.is_terminal
		or console.is_dumb_terminal
	):
		return f"{text}"
	return color_text(text, text_color)


def color_text(text: object, text_color: str = "") -> str:
	"""Return bold text in the color of your choice."""
	return (
		f"\x1b[1;{COLOR_CODES[text_color]}m{text}{COLOR_CODES['RESET']}"
		if text_color
		else f"\x1b[1m{text}{COLOR_CODES['RESET']}"
	)


def color_version(version: str) -> str:
	"""Color version number."""
	if not console.is_terminal or console.is_dumb_terminal:
		return f"({version})"
	return f"{color('(')}{color(version, 'BLUE')}{color(')')}"


translate = gettext.translation("nala", fallback=True)


def _(msg: str) -> str:
	"""Gettext translator."""
	return msg if console.options.ascii_only else translate.gettext(msg)



#                 __
#    ____ _____  |  | _____
#   /    \\__  \ |  | \__  \
#  |   |  \/ __ \|  |__/ __ \_
#  |___|  (____  /____(____  /
#       \/     \/          \/
#
# Copyright (C) 2021, 2022 Blake Lee
#
# This file is part of nala
#
# nala is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# nala is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with nala.  If not, see <https://www.gnu.org/licenses/>.
"""Nala fetch Module."""
from __future__ import annotations

import contextlib
import itertools
import re
import sys
from asyncio import Semaphore, gather, get_event_loop, run as aiorun
from pathlib import Path
from ssl import SSLCertVerificationError, SSLError
from typing import Iterable, List, Optional, Union

import typer
from apt import Cache
from apt_pkg import get_architectures
from httpx import (
	AsyncClient,
	ConnectError,
	ConnectTimeout,
	HTTPError,
	HTTPStatusError,
	Limits,
	ReadTimeout,
	Timeout,
	get,
)
from rich.progress import Progress, TaskID

from nala import _, color
from nala.constants import (
	ERROR_PREFIX,
	NALA_SOURCES,
	NOTICE_PREFIX,
	OLD_NALA_SOURCES,
	SOURCELIST,
	SOURCEPARTS,
)
from nala.downloader import print_error
from nala.error import ParserError
from nala.options import ASSUME_YES, DEBUG, MAN_HELP, VERBOSE, arguments, nala
from nala.rich import ELLIPSIS, Live, Panel, Table, fetch_progress
from nala.utils import ask, dprint, eprint, sudo_check, term

from debian.deb822 import Deb822  # isort:skip

DEBIAN = "Debian"
UBUNTU = "Ubuntu"
DEVUAN = "Devuan"

DOMAIN_PATTERN = re.compile(r"https?://([A-Za-z_0-9.-]+).*")
UBUNTU_COUNTRY = re.compile(r"<mirror:countrycode>(.*)</mirror:countrycode>")
UBUNTU_MIRROR = re.compile(r"<link>(.*)</link>")
FETCH_RANGE = re.compile(r"[0-9]+\.\.[0-9]+")
LIMITS = Limits(max_connections=50)
TIMEOUT = Timeout(timeout=5.0, read=1.0, pool=20.0)
ErrorTypes = Union[HTTPStatusError, HTTPError, SSLError, ReadTimeout, OSError]
INVALID_FILENAME_CHARS = re.compile("[^a-zA-Z0-9_.-]", re.ASCII)

FETCH_HELP = _(
	"Nala will fetch mirrors with the lowest latency.\n\n"
	"For Debian https://mirror-master.debian.org/status/Mirrors.masterlist\n\n"
	"For Ubuntu https://launchpad.net/ubuntu/+archivemirrors-rss"
)

# pylint: disable=too-many-instance-attributes
class MirrorTest:
	"""Class to test mirrors."""

	def __init__(
		self,
		netselect: tuple[str, ...],
		release: str,
		check_sources: bool,
		https_only: bool,
	):
		"""Class to test mirrors."""
		self.netselect = netselect
		self.netselect_scored: list[str] = []
		self.release = release
		self.sources = check_sources
		self.https_only = https_only
		self.client: AsyncClient
		self.progress: Progress
		self.task: TaskID

	async def run_test(self) -> None:
		"""Test mirrors."""
		with fetch_progress as self.progress:
			self.task = self.progress.add_task("", total=len(self.netselect))
			async with AsyncClient(
				follow_redirects=True, limits=LIMITS, timeout=TIMEOUT
			) as self.client:
				loop = get_event_loop()
				semp = Semaphore(25)
				tasks = [
					loop.create_task(self.net_select(mirror, semp))
					for mirror in self.netselect
				]
				await gather(*tasks)

	async def net_select(self, mirror: str, semp: Semaphore) -> None:
		"""Take a URL, ping the domain and score the latency."""
		async with semp:
			debugger = [f"Current Mirror: {mirror}"]

			regex = re.search(DOMAIN_PATTERN, mirror)
			if not regex:
				self.progress.advance(self.task)
				debugger.append("Regex Failed")
				dprint(debugger)
				return

			domain = regex[1]
			debugger.append(f"Regex Match: {domain}")
			with contextlib.suppress(RuntimeError):
				await self.netping(mirror, debugger)
			self.progress.advance(self.task)

	async def netping(self, mirror: str, debugger: list[str]) -> bool:
		"""Fetch release file and score mirror."""
		secure = False
		try:
			# Try to do https first
			https = mirror.replace("http://", "https://")
			try:
				response = await self.client.get(f"{https}dists/{self.release}/Release")
				response.raise_for_status()
				secure = True
				mirror = https
			# We catch all Exceptions because we will fall back to http
			except Exception as error:  # pylint: disable=broad-except
				debugger.append(f"https attempt failed: {error}")
				if self.https_only:
					mirror_error(error, debugger)
					dprint(debugger)
					return False

			# We can fall back to http if it's necessary
			if not secure:
				response = await self.client.get(
					f"{mirror}dists/{self.release}/Release"
				)
				response.raise_for_status()

			# Get rid of the decimal so we can prefix zeros for sorting.
			res = f"{response.elapsed.total_seconds() * 100:.0f}"
			if self.sources:
				source_response = await self.client.get(
					f"{mirror}dists/{self.release}/main/source/Release"
				)
				source_response.raise_for_status()

		# We catch all exceptions here because it really doesn't matter
		except Exception as error:  # pylint: disable=broad-except
			mirror_error(error, debugger)
			dprint(debugger)
			return False

		debugger.append(f"Download ms: {res}")
		if len(res) == 2:
			res = f"0{res}"
		elif len(res) == 1:
			res = f"00{res}"
		elif len(res) > 3:
			debugger.append("Mirror too slow")
			dprint(debugger)
			return False

		debugger.append(f"Appended: {res} {mirror}")
		dprint(debugger)
		self.netselect_scored.append(f"{res} {mirror}")
		return True

	def get_scored(self) -> tuple[str, ...]:
		"""Return sorted tuple."""
		return tuple(sorted(self.netselect_scored))


class FetchLive:
	"""Interactive Fetch."""

	def __init__(  # pylint: disable=too-many-arguments
		self,
		live: Live,
		release: str,
		sources: list[str],
		count: int,
		netselect_scored: tuple[str, ...],
	) -> None:
		"""Interactive Fetch."""
		self.live = live
		self.errors = 0
		self.count = count
		self.mirror_list: dict[int, str] = {}
		self.user_list: dict[int, str] = {}
		self.index_list: set[int]
		self._gen_mirror_list(release, sources, netselect_scored)

	def _gen_mirror_list(
		self, release: str, sources: Iterable[str], netselect_scored: Iterable[str]
	) -> None:
		"""Generate the mirror list for display."""
		index = 0
		for line in netselect_scored:
			url = line[line.index(":") :].rstrip("/")
			if any(url in mirror and release in mirror for mirror in sources):
				continue

			self.mirror_list[index] = line
			index += 1

			if len(self.mirror_list) == self.count:
				break

	def clear(self, lines: int) -> None:
		"""Clear lines for the live display."""
		for _ in range(lines + self.errors):
			term.write(term.CURSER_UP + f"\r{' '*term.columns}\r".encode())
		self.errors = 0

	def debug(self, msg: object) -> None:
		"""Display debugging information with the live display."""
		if arguments.debug:
			term.write(term.CURSER_UP * 24 + f"\r{' ' * term.columns}\r".encode())
			term.write(term.CURSER_UP + f"\r{' ' * term.columns}\r".encode())
			term.write(term.CURSER_UP + f"\r{' ' * term.columns}\r".encode())
			term.write(term.CURSER_UP + f"\r{' ' * term.columns}\r".encode())
			term.write(f"{msg}".encode())
			term.write(term.CURSER_DOWN * 26 + f"\r{' ' * term.columns}\r".encode())

	def error(self, msg: object) -> None:
		"""Print an error out and keep track of how many."""
		self.errors += 1
		term.write(term.CURSER_UP + f"\r{' ' * term.columns}\r".encode())
		eprint(msg)

	def choose_mirrors(self) -> None:
		"""Allow user to choose their mirrors."""
		while True:
			self.live.update(
				Panel.fit(
					gen_table(self.mirror_list),
					title="[bold default] Fastest Mirrors",
					title_align="left",
					border_style="bold green",
				),
				refresh=True,
			)
			self.live.stop()
			self.index_list = self.ask_index(self.count)
			if self.index_list:
				break

	def final_mirrors(self) -> bool:
		"""Confirm that the final mirrors are okay."""
		self.live.start()
		self.live.update(
			Panel.fit(
				gen_table(self.user_list),
				title="[bold white] Selected Mirrors",
				title_align="left",
				border_style="bold green",
			),
			refresh=True,
		)
		self.live.stop()
		return ask(_("Are these mirrors okay?"))

	def set_user_list(self) -> None:
		"""Set the user selected list of mirrors."""
		self.user_list = {
			num: mirror
			for num, mirror in self.mirror_list.items()
			if num in self.index_list
		}

	def ask_index(self, count: int) -> set[int]:
		"""Ask user about the mirrors they would like to use."""
		index_list: set[int] = set()
		response: str = input(
			_(
				"Mirrors you want to keep, separated by space or comma {selection}:"
			).format(selection=f"({color('1')}..{color(str(count))})")
			+ " "
		)

		# Small single use wrapper to just clean up the code a bit
		# Debug has to be done at the return or everything will get messed up
		def _debug(passthrough: set[int]) -> set[int]:
			self.debug(
				f"Response: {response or 'Default'}\n"
				f"Range: {range(1, count + 1)}\n"
				f"User Range: {passthrough}"
			)
			return passthrough

		try:
			for index in range_from_str(response, count):
				# Plus one is for taking care of the user seeing
				# numbers starting at 1 instead of 0
				if index not in range(1, count + 1):
					self.error(
						_("{error} Index {index} doesn't exist.").format(
							error=ERROR_PREFIX, index=color(index, "YELLOW")
						)
					)
					# Returning an empty set will restart the self.choose_mirrors loop
					return _debug(set())
				index_list.add(index - 1)

		except ValueError as error:
			self.error(
				_("{error} {value_error}").format(
					error=ERROR_PREFIX,
					value_error=color(f"{error}".capitalize(), "YELLOW"),
				)
			)
			# Returning an empty set will restart the self.choose_mirrors loop
			return _debug(set())

		except ParserError as error:
			self.error(
				_("{error} Parser: {parser}").format(error=ERROR_PREFIX, parser=error)
			)
		return _debug(index_list)


def mirror_error(error: Exception, debugger: list[str]) -> None:
	"""Handle errors when mirror testing."""
	if isinstance(error, HTTPStatusError):
		if arguments.verbose:
			print_error(error)
		debugger.append(f"Status Code: {error.response.status_code}")
		return

	if arguments.verbose:
		if isinstance(error, SSLCertVerificationError):
			eprint(f"{ERROR_PREFIX} {error.reason} {error.verify_message}")
		elif isinstance(error, SSLError):
			eprint(f"{ERROR_PREFIX} {error.reason}")
		elif isinstance(error, (ConnectError, ConnectTimeout)):
			print_error(error)
		else:
			eprint(f"{ERROR_PREFIX} {error}")

	if isinstance(error, ReadTimeout):
		debugger.append("Mirror too slow")


def get_and_parse_mirror(
	distro: str, country_list: Iterable[str] | None
) -> tuple[str, ...]:
	"""Get and parse the mirror list."""
	print(_("Fetching {distro} mirrors").format(distro=distro) + ELLIPSIS)
	if distro == DEBIAN:
		mirror = fetch_mirrors(
			"https://mirror-master.debian.org/status/Mirrors.masterlist", "\n\n"
		)
		# This is what one of our "Mirrors might look like after split"
		# Site: mirrors.edge.kernel.org
		# Country: NL Netherlands
		# Country: US United States
		# Location: Amsterdam
		# Location: Parsippany, NJ
		# Location: San-Jose, CA
		# Archive-architecture: amd64 arm64 armel armhf i386
		# Archive-http: /debian/
		# Sponsor: packet.net https://packet.net/
	elif distro == UBUNTU:
		mirror = fetch_mirrors(
			"https://launchpad.net/ubuntu/+archivemirrors-rss", "<item>"
		)
		# This is what one of our "Mirrors might look like after split"
		#      <title>Steadfast Networks</title>
		#      <link>http://mirror.steadfastnet.com/ubuntu/</link>
		#      <description>
		#      </description>
		#      <mirror:bandwidth>80</mirror:bandwidth>
		#      <mirror:location>
		#        <mirror:continent>North America</mirror:continent>
		#        <mirror:country>United States</mirror:country>
		#        <mirror:countrycode>US</mirror:countrycode>
		#      </mirror:location>
		#      <pubDate>Fri, 24 Dec 2021 05:26:30 -0000</pubDate>
		#      <guid>http://mirror.steadfastnet.com/ubuntu/</guid>
		#    </item>
	elif distro == DEVUAN:
		mirror = fetch_mirrors("https://pkgmaster.devuan.org/mirror_list.txt", "\n\n")
		# FQDN:  sledjhamr.org
		# BaseURL:  sledjhamr.org/devuan
		# Bandwidth:  1Gb/s
		# Rate:  30min
		# Country:  Netherlands
		# CountryCode:  NL | BE | CH | CZ | DE | DK | FR | GB | GG | IE | IM | JE | LU
		# Protocols:  HTTP | HTTPS | FTP | RSYNC
		# Active:  yes
		# DNSRR:  yes
		# DNSRRCC:  yes
	else:
		# We should never really hit this.
		sys.exit(
			_("{error} Internal Error. Distro detection must be broken").format(
				error=ERROR_PREFIX
			)
		)
	return parse_mirror(distro, mirror, country_list, tuple(get_architectures()))


def fetch_mirrors(url: str, splitter: str) -> tuple[str, ...]:
	"""Attempt to fetch the url and split a list based on the splitter."""
	try:
		response = get(url, timeout=15, follow_redirects=True)
		response.raise_for_status()
		mirror_list = response.text.split(splitter)
	except HTTPError:
		sys.exit(
			_("{error} unable to connect to {mirror}").format(
				error=ERROR_PREFIX, mirror=url
			)
		)
	return tuple(mirror_list)


def parse_mirror(
	distro: str,
	master_mirror: tuple[str, ...],
	country_list: Iterable[str] | None,
	arches: tuple[str, ...],
) -> tuple[str, ...]:
	"""Parse the mirror."""
	mirror_set = set()
	if arguments.verbose:
		print(_("Parsing mirror list") + ELLIPSIS)
	# If no country is supplied then our list will be all countries
	countries = country_list or get_countries(master_mirror)
	for country, mirror in itertools.product(countries, master_mirror):
		if (
			distro == DEBIAN
			and f"Country: {country.upper()}" in mirror
			and (url := debian_parser(mirror, arches))
		):
			mirror_set.add(url)
			continue

		if (
			distro == UBUNTU
			and f"<mirror:countrycode>{country.upper()}</mirror:countrycode>" in mirror
			and (url := ubuntu_parser(mirror, arches))
		):
			mirror_set.add(url)
			continue

		if distro == DEVUAN:
			for line in mirror.splitlines():
				# CountryCode:  NL | BE | CH
				if line.startswith("CountryCode:") and country.upper() in line:
					if url := devuan_parser(mirror):
						mirror_set.add(url)
						continue

	return tuple(mirror_set)


def get_countries(master_mirror: tuple[str, ...]) -> tuple[str, ...]:
	"""Iterate the mirror list and return all valid countries."""
	country_list = set()
	# The way we split the information we get nice and pretty mirror selections
	for mirror in master_mirror:
		for line in mirror.splitlines():
			# Devuan Countries
			if "CountryCode:" in line:
				# CountryCode:  BG | GR | RO | MK | RS | TR
				for country in line.split()[1:]:
					if line == "|":
						continue
					country_list.add(country)
			# Debian Countries
			elif "Country:" in line:
				# Country: SE Sweden
				country_list.add(line.split()[1])
			# Ubuntu Countries
			elif "<mirror:countrycode>" in line:
				# <mirror:countrycode>US</mirror:countrycode>
				if result := re.search(UBUNTU_COUNTRY, line):
					country_list.add(result[1])
	return tuple(country_list)


def devuan_parser(mirror: str) -> str | None:
	"""Parse the Debuan mirror."""
	if "HTTP" not in mirror:
		return None
	url = None
	for line in mirror.splitlines():
		# BaseURL:  sledjhamr.org/devuan
		if line.startswith("BaseURL:"):
			url = line.split()[1]

	return f"http://{url}/devuan/" if url else None


def debian_parser(mirror: str, arches: tuple[str, ...]) -> str | None:
	"""Parse the Debian mirror."""
	url = "http://"
	if "Archive-http:" in mirror and all(arch in mirror for arch in arches):
		for line in mirror.splitlines():
			if line.startswith(("Archive-http:", "Site:")):
				# ['Site:', 'mirror.steadfastnet.com']
				# ['Archive-http:', '/debian/']
				url += line.split()[1]

	return None if url == "http://" else url


def ubuntu_parser(mirror: str, arches: tuple[str, ...]) -> str | None:
	"""Parse the Ubuntu mirror."""
	# First section we get from Ubuntu is garbage. Let's ditch it and get to business
	if "<title>Ubuntu Archive Mirrors Status</title>" in mirror:
		return None
	only_ports = "amd64" not in arches and "i386" not in arches
	for line in mirror.splitlines():
		# <link>http://mirror.steadfastnet.com/ubuntu/</link>
		if result := re.search(UBUNTU_MIRROR, line):
			return None if only_ports and "ubuntu-ports" not in result[1] else result[1]
	return None


def detect_release(
	debian: str, ubuntu: str, devuan: str
) -> tuple[str | None, str | None]:
	"""Detect the distro and release."""
	# Check if the release was specified.
	for dist, switch in (
		(DEBIAN, debian),
		(DEVUAN, devuan),
		(UBUNTU, ubuntu),
	):
		if switch:
			return dist, switch

	# If no release is specified try to detect it by keyrings.
	cache = Cache()
	for keyring in (
		"devuan-keyring",
		"devuan-archive-keyring",
		"debian-archive-keyring",
		"ubuntu-keyring",
		"apt",
	):
		if (
			keyring not in cache
			or not (cand := cache[keyring].candidate)
			or not (origin := cand.origins)
		):
			continue
		return origin[0].origin, origin[0].codename

	# Something is very wrong if apt has no origin.
	# So we parse os-release to see if we can detect anything
	release_file = Path("/etc/os-release")
	if not release_file.is_file():
		# This will throw an error at the next step
		# ERROR: There was an issue detecting release.
		return None, None

	os_release: dict[str, str] = {}
	for line in release_file.read_text(encoding="utf-8", errors="replace").splitlines():
		entry = line.split("=")
		os_release[entry[0]] = entry[1].strip('"')

	# If there is no name we'll just have it throw an error
	if not (name := os_release.get("NAME")):
		return None, None

	# This block is for Debian Testing/Sid. As they don't have a codename key.
	release = os_release.get("DEBIAN_CODENAME") or os_release.get("UBUNTU_CODENAME")
	if not release and "Debian" in name:
		try:
			release = os_release["PRETTY_NAME"].split().pop()
		except IndexError:
			return name, "Unknown"
	return name, release


def parse_sources() -> list[str]:
	"""Read sources files on disk."""
	sources: list[str] = []
	for file in [*SOURCEPARTS.iterdir(), SOURCELIST]:
		if (
			file == NALA_SOURCES
			or not file.is_file()
			or INVALID_FILENAME_CHARS.search(file.name)
		):
			continue
		if file.parent == SOURCEPARTS and file.suffix not in [".list", ".sources"]:
			continue
		if file.suffix in ".sources":
			sources.extend(
				f"{deb} {uri} {suite}"
				for deb822 in Deb822.iter_paragraphs(
					file.read_text(encoding="utf-8", errors="replace")
				)
				for deb in deb822.get("Types", "").split()
				for uri in deb822.get("URIs", "").split()
				for suite in deb822.get("Suites", "").split()
				for enabled in [deb822.get("Enabled", "yes").lower()]
				if enabled not in ["no", "false"]
				and any(digit not in "0" for digit in enabled)
			)

		else:
			sources.extend(
				line
				for line in file.read_text(
					encoding="utf-8", errors="replace"
				).splitlines()
				if not line.lstrip().startswith("#") and line
			)
	return sources


def gen_table(str_list: dict[int, str]) -> Table:
	"""Generate table for the live display."""
	master_table = Table(padding=(0, 0), box=None)
	table = Table(padding=(0, 2), box=None)

	table.add_column("Index", justify="right", style="bold blue")
	table.add_column("Mirror")
	table.add_column("Score", style="bold blue")

	for num, line in str_list.items():
		latency, mirror = line.split()
		table.add_row(f"{num + 1}", mirror, f"{latency.lstrip('0')} ms")

	master_table.add_row(table)
	master_table.add_row(
		# Add in a new line and indentation to line up the text
		"\n  "
		+ _("Score is how many milliseconds it takes to download the Release file"),
		style="italic",
	)
	return master_table


def range_from_str(string: str, count: int) -> Iterable[int]:
	"""Get a range of integers from a string.

	See live.ask_index()
	"""
	# If it's empty the default is assumed
	if not string:
		# We return plus 1 for everything to match the numbers
		# That the user sees
		return list(range(1, count + 1))

	if match := FETCH_RANGE.search(string):
		resp = match.string.split("..")
		# This will mean someone is trying to get a range
		# 0..10 = 0, 1, 2, 3, and so forth
		if len(resp) == 2:
			start = int(resp[0])
			stop = int(resp[1]) + 1
			# Convert the strings from input into integers and create the range
			return range(start, stop)

		# They must be trying to get a range of even or odd
		# 0..0..10 = 0 2 4 6 8 10
		if (step := int(resp[0])) not in {0, 1}:
			raise ParserError("0 for even and 1 for odd")

		start = int(resp[1])
		stop = int(resp[2]) + 1

		is_odds = step == 1
		return {
			num
			for num in range(start, stop)
			if (is_odds and num % 2) or (not is_odds and not num % 2)
		}

	# Use a set as we don't want to have duplicate numbers
	return {int(num) for num in re.split(r",|\s", string) if num}


def build_sources(  # pylint: disable=too-many-arguments
	distro: str,
	release: str,
	component: str,
	sources: list[str],
	netselect_scored: Iterable[str],
	non_free: bool,
	fetches: int = 3,
	live: bool = False,
	check_sources: bool = False,
) -> str:
	"""Build the sources file and return it as a string."""
	source = "# Sources file built for nala\n\n"
	num = 0

	source += "Types: deb"
	if check_sources:
		source += " deb-src"
	source += "\nURIs:\n"

	for line in netselect_scored:
		# This splits off the score '030 http://mirror.steadfast.net/debian/'
		line = line[line.index("h") :]
		# This protects us from writing mirrors that we already have in the sources
		if any(line.rstrip("/") in mirror and release in mirror for mirror in sources):
			continue

		if non_free:
			try:
				get(
					f"{line}/dists/{release}/non-free-firmware/",
					timeout=15,
					follow_redirects=True,
				).raise_for_status()
			except HTTPError:
				continue

		source += f"  {line}\n"
		num += 1

		if num == fetches:
			break

	if not live and num != fetches:
		eprint(
			_("{notice} Nala was unable to fetch {num} mirrors.").format(
				notice=NOTICE_PREFIX, num=fetches
			)
		)

	source += f"Suites: {release}\n"

	source += f"Components: {component}"
	if non_free:
		source += " non-free-firmware"
	source += "\n"

	source += f"Signed-By: /usr/share/keyrings/{distro.lower()}-archive-keyring.gpg\n"

	return source


def write_sources(source: str) -> None:
	"""Write mirrors to nala-sources.list."""
	with open(NALA_SOURCES, "w", encoding="utf-8") as file:
		file.write(source)
	print(_("Sources have been written to {file}").format(file=NALA_SOURCES))

	if OLD_NALA_SOURCES.exists():
		OLD_NALA_SOURCES.unlink()


def check_supported(
	distro: str | None,
	release: str | None,
	country_list: Iterable[str] | None,
	non_free: bool,
	ctx: typer.Context,
) -> tuple[tuple[str, ...], str]:
	"""Check if the distro is supported or not.

	If the distro is supported return mirror list and component.

	Error if the distro is not supported.
	"""
	if distro and release:
		if distro in (DEBIAN, DEVUAN) and release != "n/a":
			component = "main contrib non-free" if non_free else "main"
			return get_and_parse_mirror(distro, country_list), component
		if distro == UBUNTU:
			# It's ubuntu, you probably don't care about foss
			return (
				get_and_parse_mirror(distro, country_list),
				"main restricted universe multiverse",
			)

	if distro is None or release is None:
		eprint(
			_("{error} There was an issue detecting release.").format(
				error=ERROR_PREFIX
			),
			end="\n\n",
		)
	else:
		eprint(
			_("{error} {distro} {release} is unsupported.").format(
				error=ERROR_PREFIX, distro=distro, release=release
			)
		)
	eprint(_("You can specify Ubuntu or Debian manually."), end="\n\n")
	eprint(ctx.get_help())
	sys.exit(1)


@nala.command(
	short_help=_("Fetch fast mirrors to speed up downloads."), help=FETCH_HELP
)
# pylint: disable=unused-argument,too-many-arguments,too-many-locals
def fetch(
	ctx: typer.Context,
	debian: str = typer.Option("", metavar="sid", help=_("Choose the Debian release.")),
	ubuntu: str = typer.Option(
		"", metavar="jammy", help=_("Choose the Ubuntu release.")
	),
	devuan: str = typer.Option(
		"", metavar="stable", help=_("Choose the Devuan release.")
	),
	fetches: int = typer.Option(
		0,
		help=_("Number of mirrors to fetch. [defaults: 16, --auto(3)]"),
		show_default=False,
	),
	https_only: bool = typer.Option(
		False, "--https-only", help="Only get https mirrors."
	),
	sources: bool = typer.Option(
		False, "--sources", help=_("Add the source repos for the mirrors if it exists.")
	),
	non_free: bool = typer.Option(
		False, "--non-free", help=_("Add contrib and non-free repos.")
	),
	auto: bool = typer.Option(
		False,
		"--auto",
		help=_("Run fetch uninteractively. Will still prompt for overwrite."),
	),
	debug: bool = DEBUG,
	assume_yes: bool = ASSUME_YES,
	country_list: Optional[List[str]] = typer.Option(
		None,
		"-c",
		"--country",
		metavar="US",
		help=_("Choose only mirrors of a specific ISO country code."),
	),
	verbose: bool = VERBOSE,
	man_help: bool = MAN_HELP,
) -> None:
	"""Nala will fetch mirrors with the lowest latency.

	For Debian https://mirror-master.debian.org/status/Mirrors.masterlist

	For Ubuntu https://launchpad.net/ubuntu/+archivemirrors-rss
	"""
	sudo_check()

	# Set dynamic default fetch option
	if fetches == 0:
		fetches = 3 if auto else 16

	distro, release = detect_release(debian, ubuntu, devuan)
	netselect, component = check_supported(distro, release, country_list, non_free, ctx)
	assert distro and release

	dprint(netselect)
	dprint(f"Distro: {distro}, Release: {release}, Component: {component}")

	mirror_test = MirrorTest(netselect, release, sources, https_only)
	aiorun(mirror_test.run_test())

	if not (netselect_scored := mirror_test.get_scored()):
		sys.exit(
			_("{error} Nala was unable to find any mirrors.").format(error=ERROR_PREFIX)
		)

	dprint(netselect_scored)
	dprint(f"Size of original list: {len(netselect)}")
	dprint(f"Size of scored list: {len(netselect_scored)}")
	dprint(f"Writing from: {netselect_scored[:fetches]}")

	if auto:
		source = build_sources(
			distro,
			release,
			component,
			parse_sources(),
			netselect_scored,
			non_free,
			fetches=fetches,
			check_sources=sources,
		)
		write_sources(source)
		return

	sources_list = parse_sources()
	with Live(auto_refresh=False) as live:
		fetch_live = FetchLive(live, release, sources_list, fetches, netselect_scored)
		while True:
			fetch_live.choose_mirrors()
			fetch_live.clear(2)
			fetch_live.set_user_list()
			if fetch_live.final_mirrors():
				break
			fetch_live.clear(2)
			fetch_live.live.start()
		source = build_sources(
			distro,
			release,
			component,
			sources_list,
			fetch_live.user_list.values(),
			non_free,
			live=True,
			check_sources=sources,
		)
	write_sources(source)
